# type: ignore
from sys import platform
import unittest
from unittest.mock import patch
from pathlib import Path
import tempfile
import shutil

from datetime import datetime, timezone

import pytest

from safety.tool.interceptors.unix import UnixAliasInterceptor


@pytest.mark.unix_only
@pytest.mark.skipif(
    platform not in ["linux", "linux2", "darwin"], reason="Unix-specific tests"
)
class TestUnixAliasInterceptor(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)

        self.mock_config_patcher = patch(
            "safety.tool.interceptors.unix.USER_CONFIG_DIR",
            Path(self.temp_dir) / ".safety",
        )
        self.mock_config_patcher.start()

        self.mock_home_patcher = patch("safety.tool.interceptors.unix.Path.home")
        self.mock_datetime_patcher = patch("safety.tool.interceptors.base.datetime")
        self.mock_version_patcher = patch("safety.tool.interceptors.base.get_version")

        self.mock_home = self.mock_home_patcher.start()
        self.mock_datetime = self.mock_datetime_patcher.start()
        self.mock_version = self.mock_version_patcher.start()

        self.mock_home.return_value = Path(self.temp_dir)
        self.mock_version.return_value = "1.0.0"
        self.mock_now = datetime(2024, 1, 1, tzinfo=timezone.utc)
        self.mock_datetime.now.return_value = self.mock_now

        self.addCleanup(self.mock_home_patcher.stop)
        self.addCleanup(self.mock_datetime_patcher.stop)
        self.addCleanup(self.mock_version_patcher.stop)
        self.addCleanup(self.mock_config_patcher.stop)

    def test_interceptors_all_tools(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors()

        self.assertTrue(result)

        rc_files = (".zshrc", ".bashrc", ".bash_profile")

        rc_paths = [Path(self.temp_dir) / rc for rc in rc_files]
        safety_profile_path = Path(self.temp_dir) / ".safety" / ".safety_profile"

        self.assertTrue(
            all(path.exists() for path in rc_paths), "Not all rc_paths exist"
        )
        self.assertTrue(safety_profile_path.exists())

        expected_rc_content = (
            "# >>> Safety >>>\n"
            f'[ -f "{safety_profile_path}" ] && . "{safety_profile_path}"\n'
            "# <<< Safety <<<\n"
        )

        expected_safety_profile_content = (
            "# >>> Safety >>>\n"
            "# DO NOT EDIT THIS FILE DIRECTLY\n"
            f"# Last updated at: {self.mock_now.isoformat()}\n"
            "# Updated by: safety v1.0.0\n"
            'alias pip="safety pip"\n'
            'alias pip3="safety pip3"\n'
            'alias pip3.8="safety pip3.8"\n'
            'alias pip3.9="safety pip3.9"\n'
            'alias pip3.10="safety pip3.10"\n'
            'alias pip3.11="safety pip3.11"\n'
            'alias pip3.12="safety pip3.12"\n'
            'alias pip3.13="safety pip3.13"\n'
            'alias pip3.14="safety pip3.14"\n'
            'alias poetry="safety poetry"\n'
            'alias uv="safety uv"\n'
            "# <<< Safety <<<\n"
        )

        for rc_path in rc_paths:
            self.assertEqual(rc_path.read_text(), expected_rc_content)

        self.assertEqual(
            safety_profile_path.read_text(), expected_safety_profile_content
        )

        result = interceptor.remove_interceptors()

        self.assertTrue(result)
        self.assertTrue(
            all(path.exists() for path in rc_paths), "Some rc_paths were removed"
        )

        for rc_path in rc_paths:
            self.assertEqual(rc_path.read_text(), "")

        self.assertFalse(safety_profile_path.exists())

    def test_install_interceptors_nonexistent_tool(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors(["nonexistent"])
        self.assertFalse(result)

    def test_uninstall_interceptors_all_tools(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.remove_interceptors()
        self.assertTrue(result)



# type: ignore

import pytest
from unittest.mock import MagicMock, patch
from typing import List, Tuple, Optional

import typer

from safety.models import ToolResult
from safety.tool.mixins import InstallationAuditMixin
from safety.tool.environment_diff import EnvironmentDiffTracker


class MockAuditableCommand(InstallationAuditMixin):
    """
    Mock implementation of an auditable command.
    """

    def __init__(self, packages=None, diff_data=None):
        self._mock_packages = packages or []
        self._mock_diff_data = diff_data or ({}, {}, {})

    @property
    def _diff_tracker(self) -> "EnvironmentDiffTracker":
        mock_tracker = MagicMock(spec=EnvironmentDiffTracker)
        mock_tracker.get_diff.return_value = self._mock_diff_data
        return mock_tracker

    @property
    def _packages(self) -> List[Tuple[str, Optional[str]]]:
        return self._mock_packages


class TestInstallationAuditMixin:
    """
    Test suite for InstallationAuditMixin functionality.
    """

    def setup_method(self):
        """
        Set up test fixtures.
        """
        self.ctx = MagicMock(spec=typer.Context)
        self.ctx.obj = MagicMock()
        self.ctx.obj.auth = MagicMock()
        self.ctx.obj.auth.client = MagicMock()
        self.result = MagicMock(spec=ToolResult)
        self.result.process = MagicMock()
        self.result.process.returncode = 0
        self.result.duration_ms = 100  # Add the missing duration_ms attribute

    @patch("safety.tool.mixins.console")
    def test_audit_packages_with_diff(self, mock_console):
        """
        Test audit_packages method with diff tracker.
        """
        added_packages = {"package1": "1.0.0", "package2": "2.0.0"}
        updated_packages = {"package3": "3.0.0"}
        diff_data = (added_packages, {}, updated_packages)  # (added, removed, updated)

        command = MockAuditableCommand(diff_data=diff_data)
        self.ctx.obj.auth.client.audit_packages.return_value = {
            "audit": {
                "packages": [
                    {
                        "package_specifier": "package1==1.0.0",
                        "vulnerabilities": {"critical": 2, "high": 1},
                    }
                ]
            }
        }

        result = command.audit_packages(self.ctx)

        self.ctx.obj.auth.client.audit_packages.assert_called_once()
        expected_packages = ["package1==1.0.0", "package2==2.0.0", "package3==3.0.0"]
        actual_packages = self.ctx.obj.auth.client.audit_packages.call_args[0][0]
        assert sorted(actual_packages) == sorted(expected_packages)
        assert result == self.ctx.obj.auth.client.audit_packages.return_value

    @patch("safety.tool.mixins.console")
    def test_audit_packages_with_tuple_version(self, mock_console):
        """
        Test audit_packages method with tuple version.
        """
        added_packages = {"package1": ("0.9.0", "1.0.0")}
        diff_data = (added_packages, {}, {})

        command = MockAuditableCommand(diff_data=diff_data)
        self.ctx.obj.auth.client.audit_packages.return_value = {
            "audit": {"packages": []}
        }

        result = command.audit_packages(self.ctx)

        self.ctx.obj.auth.client.audit_packages.assert_called_once_with(
            ["package1==1.0.0"]
        )
        assert result == self.ctx.obj.auth.client.audit_packages.return_value

    @patch("safety.tool.mixins.console")
    def test_audit_packages_without_auth(self, mock_console):
        """
        Test audit_packages method without auth client.
        """
        command = MockAuditableCommand()
        self.ctx.obj.auth = None

        result = command.audit_packages(self.ctx)

        assert result == {}

    @patch("safety.tool.mixins.console")
    def test_audit_packages_with_exception(self, mock_console):
        """
        Test audit_packages method with exception.
        """
        command = MockAuditableCommand(diff_data=({"package1": "1.0.0"}, {}, {}))
        self.ctx.obj.auth.client.audit_packages.side_effect = Exception("API error")

        result = command.audit_packages(self.ctx)

        assert result == {}

    @patch("safety.tool.mixins.console")
    @patch("safety.tool.mixins.render_header")
    @patch("safety.tool.mixins.progressive_print")
    def test_render_installation_warnings_with_vulnerabilities(
        self, mock_progressive_print, mock_render_header, mock_console
    ):
        """
        Test render_installation_warnings with vulnerability data.
        """
        command = MockAuditableCommand()
        packages_audit = {
            "audit": {
                "packages": [
                    {
                        "package_specifier": "vulnerable-package==1.0.0",
                        "vulnerabilities": {"critical": 2, "high": 1, "medium": 3},
                    },
                    {
                        "package_specifier": "safe-package==2.0.0",
                        "vulnerabilities": {},  # A backend mismatch
                    },
                ]
            }
        }

        command.render_installation_warnings(self.ctx, packages_audit)

        mock_render_header.assert_called_once_with(" Safety Report")
        mock_progressive_print.assert_called_once()
        warning_messages = mock_progressive_print.call_args[0][0]
        assert len(warning_messages) == 1
        assert (
            "vulnerable-package==1.0.0 contains 6 vulnerabilities"
            in warning_messages[0]
        )
        assert "including 2 critical severity vulnerabilities" in warning_messages[0]

    @patch("safety.tool.mixins.console")
    def test_render_installation_warnings_without_audit_data(self, mock_console):
        """
        Test render_installation_warnings without audit data.
        """
        command = MockAuditableCommand()
        command.render_installation_warnings(self.ctx, {})
        mock_console.print.assert_not_called()

    @patch("safety.tool.mixins.console")
    def test_render_package_details(self, mock_console):
        """
        Test render_package_details method.
        """
        packages = [("package1", "1.0.0"), ("package2", None)]
        command = MockAuditableCommand(packages=packages)

        command.render_package_details()

        assert mock_console.print.call_count == 2
        for call_args in mock_console.print.call_args_list:
            padding_arg = call_args[0][0]
            assert "Learn more: " in str(padding_arg)
            assert "https://data.safetycli.com/packages/pypi/" in str(padding_arg)

    @patch.object(MockAuditableCommand, "audit_packages")
    @patch.object(MockAuditableCommand, "render_installation_warnings")
    @patch.object(MockAuditableCommand, "render_package_details")
    @patch("safety.tool.base.BaseCommand._perform_diff")
    @patch("safety.tool.base.BaseCommand._handle_command_result")
    def test_handle_installation_audit_success(
        self,
        mock_handle_result,
        mock_perform_diff,
        mock_render_details,
        mock_render_warnings,
        mock_audit_packages,
    ):
        """
        Test handle_installation_audit with successful process.
        """
        command = MockAuditableCommand()
        mock_audit_packages.return_value = {"audit": {"packages": []}}
        self.result.process.returncode = 0

        command.handle_installation_audit(self.ctx, self.result)

        mock_audit_packages.assert_called_once_with(self.ctx)
        mock_render_warnings.assert_called_once_with(
            self.ctx, mock_audit_packages.return_value
        )
        mock_render_details.assert_not_called()  # Should not be called for successful process

    @patch.object(MockAuditableCommand, "audit_packages")
    @patch.object(MockAuditableCommand, "render_installation_warnings")
    @patch.object(MockAuditableCommand, "render_package_details")
    @patch("safety.tool.base.BaseCommand._perform_diff")
    @patch("safety.tool.base.BaseCommand._handle_command_result")
    def test_handle_installation_audit_failure(
        self,
        mock_handle_result,
        mock_perform_diff,
        mock_render_details,
        mock_render_warnings,
        mock_audit_packages,
    ):
        """
        Test handle_installation_audit with failed process.
        """

        command = MockAuditableCommand()
        mock_audit_packages.return_value = {"audit": {"packages": []}}
        self.result.process.returncode = 1  # Non-zero indicates failure

        command.handle_installation_audit(self.ctx, self.result)

        mock_audit_packages.assert_called_once_with(self.ctx)
        mock_render_warnings.assert_called_once_with(
            self.ctx, mock_audit_packages.return_value
        )
        mock_render_details.assert_called_once()  # Should be called for failed process

    def test_handle_installation_audit_type_error(self):
        """
        Test handle_installation_audit with non-AuditableCommand instance.
        """
        non_auditable = InstallationAuditMixin()

        with pytest.raises(TypeError):
            non_auditable.handle_installation_audit(self.ctx, self.result)



import tempfile
import os
from pathlib import Path
from unittest.mock import patch

from safety.encoding import detect_encoding


class TestDetectEncoding:
    """
    Tests for the detect_encoding function.
    """

    def test_utf16_le_detection(self):
        """
        Test that UTF-16 LE (Little Endian) encoding is correctly detected.
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            # UTF-16 LE BOM: \xff\xfe
            f.write(b"\xff\xfe" + "Hello, world!".encode("utf-16-le")[2:])

        try:
            encoding = detect_encoding(Path(f.name))
            assert encoding == "utf-16"
        finally:
            os.unlink(f.name)

    def test_utf16_be_detection(self):
        """
        Test that UTF-16 BE (Big Endian) encoding is correctly detected.
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            # UTF-16 BE BOM: \xfe\xff
            f.write(b"\xfe\xff" + "Hello, world!".encode("utf-16-be")[2:])

        try:
            encoding = detect_encoding(Path(f.name))
            assert encoding == "utf-16"
        finally:
            os.unlink(f.name)

    def test_utf8_sig_detection(self):
        """
        Test that UTF-8 with signature (BOM) is correctly detected.
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            # UTF-8 with signature BOM: \xef\xbb\xbf
            f.write(b"\xef\xbb\xbf" + "Hello, world!".encode("utf-8"))

        try:
            encoding = detect_encoding(Path(f.name))
            assert encoding == "utf-8-sig"
        finally:
            os.unlink(f.name)

    def test_utf8_detection(self):
        """
        Test that regular UTF-8 (without BOM) is correctly detected.
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            # Regular UTF-8 (no BOM)
            f.write("Hello, world!".encode("utf-8"))

        try:
            encoding = detect_encoding(Path(f.name))
            assert encoding == "utf-8"
        finally:
            os.unlink(f.name)

    def test_error_handling(self):
        """
        Test that errors are properly handled and default encoding is returned.
        """
        # Test with a non-existent file
        non_existent_file = Path("non_existent_file.txt")

        encoding = detect_encoding(non_existent_file)
        assert encoding == "utf-8"

    def test_exception_logging(self):
        """
        Test that exceptions are properly logged.
        """
        non_existent_file = Path("non_existent_file.txt")

        with patch("safety.encoding.logger") as mock_logger:
            encoding = detect_encoding(non_existent_file)

            assert encoding == "utf-8"
            mock_logger.exception.assert_called_once_with("Error detecting encoding")



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from rich.console import Console


def render_initialization_result(
    console: "Console",
    codebase_init_status: Optional[str] = None,
    codebase_id: Optional[str] = None,
):
    if not codebase_init_status or not codebase_id:
        console.print("Error: unable to initialize codebase")
        return

    message = None

    if codebase_init_status == "created":
        from safety.codebase.constants import CODEBASE_INIT_CREATED

        message = CODEBASE_INIT_CREATED

    if codebase_init_status == "linked":
        from safety.codebase.constants import CODEBASE_INIT_LINKED

        message = CODEBASE_INIT_LINKED

    if codebase_init_status == "reinitialized":
        from safety.codebase.constants import CODEBASE_INIT_REINITIALIZED

        message = CODEBASE_INIT_REINITIALIZED

    if message:
        console.print(message.format(codebase_name=codebase_id))



from typing import TYPE_CHECKING, Optional
from ..codebase_utils import load_unverified_project_from_config
from safety.errors import SafetyError, SafetyException
from pathlib import Path
from safety.codebase.constants import (
    CODEBASE_INIT_ERROR,
    CODEBASE_INIT_NOT_FOUND_LINK_TO,
    CODEBASE_INIT_NOT_FOUND_PROJECT_FILE,
)
from safety.init.main import create_project
from typer import Context
from rich.console import Console
import sys

if TYPE_CHECKING:
    from ..codebase_utils import UnverifiedProjectModel


def initialize_codebase(
    ctx: Context,
    console: Console,
    codebase_path: Path,
    unverified_codebase: "UnverifiedProjectModel",
    org_slug: str,
    link_to: Optional[str] = None,
    should_enable_firewall: bool = False,
):
    is_interactive = sys.stdin.isatty()
    link_behavior = "prompt"
    create_if_missing = True
    is_codebase_file_created = unverified_codebase.created

    if link_to or is_codebase_file_created:
        link_behavior = "always"
        create_if_missing = False
    elif not is_interactive:
        link_behavior = "never"

    project_file_created, project_status = create_project(
        ctx=ctx,
        console=console,
        target=codebase_path,
        unverified_project=unverified_codebase,
        create_if_missing=create_if_missing,
        link_behavior=link_behavior,
    )

    if project_status == "not_found":
        codebase_name = "Unknown"
        msg = "Codebase not found."

        if link_to:
            msg = CODEBASE_INIT_NOT_FOUND_LINK_TO
            codebase_name = link_to
        elif is_codebase_file_created:
            msg = CODEBASE_INIT_NOT_FOUND_PROJECT_FILE
            codebase_name = unverified_codebase.id

        raise SafetyError(msg.format(codebase_name=codebase_name))
    elif project_status == "found" and not is_interactive:
        # Non-TTY mode: Project exists but we can't link (link_behavior="never")
        suggested_name = unverified_codebase.id
        raise SafetyError(
            f"Project '{suggested_name}' already exists. "
            f"In non-interactive mode, use --link-to '{suggested_name}' to link to the existing project, "
            f"or use --name with a different project name to create a new one."
        )

    if not ctx.obj.project:
        raise SafetyException(CODEBASE_INIT_ERROR)

    if should_enable_firewall:
        from ..tool.main import configure_local_directory

        configure_local_directory(codebase_path, org_slug, ctx.obj.project.id)

    return project_file_created, project_status


def fail_if_codebase_name_mismatch(
    provided_name: str,
    unverified_codebase: "UnverifiedProjectModel",
) -> None:
    """
    Useful to prevent the user from overwriting an existing codebase by mistyping the name.
    """
    if unverified_codebase.id and provided_name != unverified_codebase.id:
        from safety.codebase.constants import CODEBASE_INIT_ALREADY_EXISTS

        raise SafetyError(CODEBASE_INIT_ALREADY_EXISTS)


def prepare_unverified_codebase(
    codebase_path: Path,
    user_provided_name: Optional[str] = None,
    user_provided_link_to: Optional[str] = None,
) -> "UnverifiedProjectModel":
    """
    Prepare the unverified codebase object based on the provided name and link to.
    """
    unverified_codebase = load_unverified_project_from_config(
        project_root=codebase_path
    )

    provided_name = user_provided_name or user_provided_link_to

    if provided_name:
        fail_if_codebase_name_mismatch(
            provided_name=provided_name,
            unverified_codebase=unverified_codebase,
        )

        unverified_codebase.id = provided_name

    return unverified_codebase



import pytest
from typing import List, Tuple, Generator
from unittest.mock import Mock, MagicMock, patch
from pathlib import Path

from safety.tool.interceptors.windows import AutoRunManager


@pytest.fixture(scope="module", autouse=True)
def ensure_winreg_mockable():
    """
    Ensure winreg can be mocked on all platforms by making it available
    as an attribute on the windows module if it doesn't exist.
    """
    from safety.tool.interceptors import windows

    # If winreg doesn't exist as an attribute, create a placeholder
    if not hasattr(windows, "winreg"):
        windows.winreg = None

    yield

    # Clean up if we added it
    if hasattr(windows, "winreg") and windows.winreg is None:
        delattr(windows, "winreg")


TOKENIZER_TEST_CASES = [
    # Basic cases
    ("", []),
    ("doskey.exe", ["doskey.exe"]),
    ("cmd1 & cmd2", ["cmd1 ", "&", " cmd2"]),
    ("cmd1&&cmd2", ["cmd1", "&&", "cmd2"]),
    ("cmd1 || cmd2", ["cmd1 ", "||", " cmd2"]),
    (
        "cmd1 & cmd2 && cmd3 | cmd4 || cmd5",
        ["cmd1 ", "&", " cmd2 ", "&&", " cmd3 ", "|", " cmd4 ", "||", " cmd5"],
    ),
    # Quoted strings
    (
        '"C:\\Program Files\\Tool\\app.exe" && script.bat',
        ['"C:\\Program Files\\Tool\\app.exe" ', "&&", " script.bat"],
    ),
    # Whitespace variations
    ("cmd1&&cmd2||cmd3", ["cmd1", "&&", "cmd2", "||", "cmd3"]),
    ("cmd1  &  cmd2", ["cmd1  ", "&", "  cmd2"]),
    ("  cmd1 & cmd2  ", ["  cmd1 ", "&", " cmd2  "]),
    # Doskey scenarios
    ("doskey ls=dir $*", ["doskey ls=dir $*"]),
    (
        'doskey /macrofile="C:\\aliases.txt" && echo Loaded',
        ['doskey /macrofile="C:\\aliases.txt" ', "&&", " echo Loaded"],
    ),
    # Corporate scenarios
    (
        '"C:\\McAfee\\agent.exe" --background && "C:\\Scripts\\init.bat"',
        ['"C:\\McAfee\\agent.exe" --background ', "&&", ' "C:\\Scripts\\init.bat"'],
    ),
    (
        "antivirus.exe /silent && logger.exe --start && cleanup.bat",
        ["antivirus.exe /silent ", "&&", " logger.exe --start ", "&&", " cleanup.bat"],
    ),
    # Environment variables
    (
        '"%USERPROFILE%\\Scripts\\startup.bat"',
        ['"%USERPROFILE%\\Scripts\\startup.bat"'],
    ),
    (
        '"%ProgramFiles%\\Tool\\app.exe" && script.bat',
        ['"%ProgramFiles%\\Tool\\app.exe" ', "&&", " script.bat"],
    ),
    # Edge cases
    ("& script.bat", ["&", " script.bat"]),
    ("script.bat &", ["script.bat ", "&"]),
    ("&", ["&"]),
    ("&&", ["&&"]),
    # Path variations
    (
        "C:\\PROGRA~1\\Tool\\app.exe && script.bat",
        ["C:\\PROGRA~1\\Tool\\app.exe ", "&&", " script.bat"],
    ),
    (
        '"C:/Scripts/startup.bat" && echo Done',
        ['"C:/Scripts/startup.bat" ', "&&", " echo Done"],
    ),
    # Parameters
    (
        "tool.exe /silent && script.bat --verbose",
        ["tool.exe /silent ", "&&", " script.bat --verbose"],
    ),
    (
        'app.exe --config="C:\\config.ini" && next.bat',
        ['app.exe --config="C:\\config.ini" ', "&&", " next.bat"],
    ),
    # Complex realistic scenarios
    (
        '"C:\\Company\\Security\\endpoint.exe" --silent && "C:\\Scripts\\inventory.bat" && echo Done',
        [
            '"C:\\Company\\Security\\endpoint.exe" --silent ',
            "&&",
            ' "C:\\Scripts\\inventory.bat" ',
            "&&",
            " echo Done",
        ],
    ),
]


@pytest.mark.unit
class TestAutoRunManagerLogicMethods:
    """
    Test pure logic methods that don't touch registry
    """

    @pytest.mark.parametrize("input_val,expected", TOKENIZER_TEST_CASES)
    def test_tokenize_autorun_scenarios(
        self, input_val: str, expected: List[str]
    ) -> None:
        manager = AutoRunManager()
        result = manager._tokenize_autorun(input_val)
        assert result == expected

    @pytest.mark.parametrize(
        "separator,expected",
        [
            ("&", True),
            ("&&", True),
            ("|", True),
            ("||", True),
            ("  &  ", True),
            ("command", False),
            ("", False),
            ("cmd&script", False),
        ],
    )
    def test_separator_detection(self, separator: str, expected: bool) -> None:
        manager = AutoRunManager()
        result = manager._is_separator(separator)
        assert result == expected

    def test_script_existence_detection(self) -> None:
        manager = AutoRunManager()
        tokens = ["script1.bat", " & ", "script2.bat", "  script3.bat  "]

        assert manager._script_exists_in_tokens(tokens, "script1.bat")
        assert manager._script_exists_in_tokens(
            tokens, "script3.bat"
        )  # whitespace ignored
        assert not manager._script_exists_in_tokens(tokens, "nonexistent.bat")
        assert not manager._script_exists_in_tokens([], "any_script.bat")

    def test_script_token_removal(self) -> None:
        manager = AutoRunManager()

        tokens = ["script1.bat", " & ", "script2.bat"]
        result = manager._remove_script_tokens(tokens, "script1.bat")
        assert result == ["script2.bat"]

        tokens = ["script1.bat", "script2.bat"]
        result = manager._remove_script_tokens(tokens, "script1.bat")
        assert result == ["script2.bat"]

        tokens = ["cmd1", " & ", "script.bat", " && ", "cmd2"]
        result = manager._remove_script_tokens(tokens, "script.bat")
        assert result == ["cmd1", "&", "cmd2"]

        tokens = ["script1.bat", "&", "script2.bat"]
        result = manager._remove_script_tokens(tokens, "nonexistent.bat")
        assert result == tokens


@pytest.mark.unit
class TestAutoRunManagerRegistryOperations:
    """
    Test AutoRun registry operations with mocked winreg
    """

    @pytest.fixture
    def mock_winreg(self) -> Generator[Tuple[Mock, Mock], None, None]:
        """
        Fixture providing mocked winreg module and key
        """
        with patch("safety.tool.interceptors.windows.winreg") as mock_reg:
            mock_key = MagicMock()
            mock_reg.OpenKey.return_value.__enter__.return_value = mock_key
            mock_reg.CreateKey.return_value.__enter__.return_value = mock_key
            mock_reg.HKEY_CURRENT_USER = "HKEY_CURRENT_USER"
            mock_reg.KEY_READ = 1
            mock_reg.KEY_SET_VALUE = 2
            mock_reg.REG_SZ = 1
            yield mock_reg, mock_key

    def test_registry_key_operations(self, mock_winreg: Tuple[Mock, Mock]) -> None:
        """Test registry key opening and creation"""
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test successful key opening
        with manager._open_registry_key() as key:
            assert key == mock_key
        mock_reg.OpenKey.assert_called_once()

        # Test key creation when not found
        mock_reg.OpenKey.side_effect = FileNotFoundError()
        mock_reg.reset_mock()

        with manager._open_registry_key() as key:
            assert key == mock_key
        mock_reg.CreateKey.assert_called_once()

    def test_get_current_tokens_operations(
        self, mock_winreg: Tuple[Mock, Mock]
    ) -> None:
        """Test getting current tokens from registry"""
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test with existing value
        mock_reg.QueryValueEx.return_value = ("cmd1 & cmd2", mock_reg.REG_SZ)
        result = manager._get_current_tokens(mock_key)
        assert result == ["cmd1 ", "&", " cmd2"]

        # Test with missing value
        mock_reg.QueryValueEx.side_effect = FileNotFoundError()
        result = manager._get_current_tokens(mock_key)
        assert result == []

    def test_registry_value_modification(self, mock_winreg: Tuple[Mock, Mock]) -> None:
        """Test setting and deleting registry values"""
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test setting value
        test_value = "test_script.bat & other_cmd"
        manager._set_autorun_value(mock_key, test_value)
        mock_reg.SetValueEx.assert_called_once_with(
            mock_key, "AutoRun", 0, mock_reg.REG_SZ, test_value
        )

        # Test deleting value
        manager._delete_autorun_value(mock_key)
        mock_reg.DeleteValue.assert_called_once_with(mock_key, "AutoRun")


@pytest.mark.unit
class TestAutoRunManagerPublicAPI:
    """
    Test the public API methods (add_script, remove_script, get_current_commands)
    """

    @pytest.fixture
    def mock_winreg(self) -> Generator[Tuple[Mock, Mock], None, None]:
        """Fixture for public API tests"""
        with patch("safety.tool.interceptors.windows.winreg") as mock_reg:
            mock_key = MagicMock()
            mock_reg.OpenKey.return_value.__enter__.return_value = mock_key
            mock_reg.REG_SZ = 1
            yield mock_reg, mock_key

    def test_add_script_scenarios(self, mock_winreg: Tuple[Mock, Mock]) -> None:
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test adding to empty AutoRun
        mock_reg.QueryValueEx.side_effect = FileNotFoundError()
        result = manager.add_script("test_script.bat")
        assert result is True
        mock_reg.SetValueEx.assert_called_with(
            mock_key, "AutoRun", 0, mock_reg.REG_SZ, "test_script.bat"
        )

        # Test adding to existing AutoRun
        mock_reg.reset_mock()
        mock_reg.QueryValueEx.side_effect = None
        mock_reg.QueryValueEx.return_value = ("existing_cmd", mock_reg.REG_SZ)
        result = manager.add_script("new_script.bat")
        assert result is True
        mock_reg.SetValueEx.assert_called_with(
            mock_key, "AutoRun", 0, mock_reg.REG_SZ, "new_script.bat & existing_cmd"
        )

        # Test adding duplicate (should not modify)
        mock_reg.reset_mock()
        mock_reg.QueryValueEx.return_value = (
            "existing_cmd & test_script.bat",
            mock_reg.REG_SZ,
        )
        result = manager.add_script("test_script.bat")
        assert result is True
        mock_reg.SetValueEx.assert_not_called()

        # Test Path object conversion
        mock_reg.reset_mock()
        mock_reg.QueryValueEx.side_effect = FileNotFoundError()
        script_path = Path("C:\\test\\script.bat")
        result = manager.add_script(script_path)
        assert result is True
        mock_reg.SetValueEx.assert_called_with(
            mock_key, "AutoRun", 0, mock_reg.REG_SZ, str(script_path)
        )

    def test_remove_script_scenarios(self, mock_winreg: Tuple[Mock, Mock]) -> None:
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test removing from multiple commands
        mock_reg.QueryValueEx.return_value = (
            "test_script.bat & other_cmd",
            mock_reg.REG_SZ,
        )
        result = manager.remove_script("test_script.bat")
        assert result is True
        mock_reg.SetValueEx.assert_called_with(
            mock_key, "AutoRun", 0, mock_reg.REG_SZ, "other_cmd"
        )

        # Test removing only command (should delete value)
        mock_reg.reset_mock()
        mock_reg.QueryValueEx.return_value = ("test_script.bat", mock_reg.REG_SZ)
        result = manager.remove_script("test_script.bat")
        assert result is True
        mock_reg.DeleteValue.assert_called_once_with(mock_key, "AutoRun")
        mock_reg.SetValueEx.assert_not_called()

        # Test removing non-existent script
        mock_reg.reset_mock()
        mock_reg.QueryValueEx.return_value = ("other_cmd", mock_reg.REG_SZ)
        result = manager.remove_script("nonexistent.bat")
        assert result is True
        mock_reg.SetValueEx.assert_not_called()
        mock_reg.DeleteValue.assert_not_called()

    def test_get_current_commands_scenarios(
        self, mock_winreg: Tuple[Mock, Mock]
    ) -> None:
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test with commands and separators
        autorun_value = "  cmd1.bat  & cmd2.exe && cmd3.bat  "
        mock_reg.QueryValueEx.return_value = (autorun_value, mock_reg.REG_SZ)
        result = manager.get_current_commands()
        assert result == ["cmd1.bat", "cmd2.exe", "cmd3.bat"]

        # Test with empty/whitespace tokens
        autorun_value = "cmd1.bat &  & cmd2.exe &   & cmd3.bat"
        mock_reg.QueryValueEx.return_value = (autorun_value, mock_reg.REG_SZ)
        result = manager.get_current_commands()
        assert result == ["cmd1.bat", "cmd2.exe", "cmd3.bat"]

        # Test with no AutoRun value
        mock_reg.QueryValueEx.side_effect = FileNotFoundError()
        result = manager.get_current_commands()
        assert result == []

    def test_error_handling(self, mock_winreg: Tuple[Mock, Mock]) -> None:
        mock_reg, mock_key = mock_winreg
        manager = AutoRunManager()

        # Test add_script error handling
        mock_reg.OpenKey.side_effect = Exception("Registry error")
        result = manager.add_script("test.bat")
        assert result is False

        # Test remove_script error handling
        result = manager.remove_script("test.bat")
        assert result is False

        # Test get_current_commands error handling
        result = manager.get_current_commands()
        assert result == []


@pytest.mark.unit
class TestAutoRunManagerEdgeCases:
    def test_edge_cases_and_defensive_programming(self) -> None:
        manager = AutoRunManager()

        # Test with empty script path
        tokens = ["script.bat", " & ", "other.bat"]
        expected_tokens = [token.strip() for token in tokens]

        assert not manager._script_exists_in_tokens(tokens, "")
        assert manager._remove_script_tokens(tokens, "") == expected_tokens

        assert manager._tokenize_autorun(None) == []  # type: ignore

        with pytest.raises(AttributeError):
            manager._is_separator(None)  # type: ignore



from importlib.metadata import distributions
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def get_project_dependencies():
    try:
        with open("pyproject.toml", "rb") as f:
            pyproject = tomllib.load(f)
            deps = pyproject.get("project", {}).get("dependencies", [])
            return {
                canonicalize_name(Requirement(dep).name): dep
                for dep in deps
                if isinstance(dep, str)
            }
    except Exception as e:
        print(f"Error reading dependencies: {e}")
        return {}


def pytest_configure(config):
    main_deps_specs = get_project_dependencies()
    all_dists = {
        canonicalize_name(dist.metadata["Name"]): (dist.metadata["Name"], dist.version)
        for dist in distributions()
    }

    # Main dependencies table
    print(f"\n[{len(main_deps_specs)}] Main Dependencies:")
    print("-" * 60)
    print("%-20s %-25s %-15s" % ("Package", "Specification", "Installed"))
    print("-" * 60)

    for pkg_norm, spec in sorted(main_deps_specs.items()):
        if pkg_norm in all_dists:
            name, version = all_dists[pkg_norm]
            print("%-20s %-25s %-15s" % (name, spec, version))

    other_pkgs = [
        f"{name} {ver}"
        for pkg_norm, (name, ver) in sorted(all_dists.items())
        if pkg_norm not in main_deps_specs
    ]

    # Other dependencies in wrapped format
    print(f"\n[{len(other_pkgs)}] Other Dependencies:")
    print("-" * 80)

    # Print other dependencies with wrapping
    line = ""
    for pkg in other_pkgs:
        if len(line) + len(pkg) + 2 > 78:
            print(line.rstrip(", "))
            line = pkg + ", "
        else:
            line += pkg + ", "
    if line:
        print(line.rstrip(", "))



# type: ignore
import unittest

from safety.tool.poetry.command import PoetryCommand
from ..base import CommandToolCaseMixin


class TestPoetryCommand(CommandToolCaseMixin, unittest.TestCase):
    """
    Test cases for PoetryCommand functionality.
    """

    command_class = PoetryCommand
    command_args = ["poetry", "add", "foobar"]
    parent_env = {"EXISTING_VAR": "existing_value", "PATH": "/usr/bin:/bin"}
    expected_env_vars = {
        "POETRY_HTTP_BASIC_SAFETY_USERNAME": "user",
        "POETRY_HTTP_BASIC_SAFETY_PASSWORD": "mock_credentials_value",
    }
    mock_configurations = [
        {
            "target": "safety.tool.poetry.command.index_credentials",
            "return_value": "mock_credentials_value",
        }
    ]



# Custom emoji namespace mapping
import re
from typing import Match


CUSTOM_EMOJI_MAP = {
    "icon_check": "✓",
    "icon_warning": "⚠️",
    "icon_info": "ℹ️",
}

# ASCII fallback mapping for problematic environments
ASCII_FALLBACK_MAP = {
    "icon_check": "+",
    "icon_warning": "!",
    "icon_info": "i",
    "white_heavy_check_mark": "++",
    "white_check_mark": "+",
    "check_mark": "+",
    "heavy_check_mark": "+",
    "shield": "[SHIELD]",
    "x": "X",
    "lock": "[LOCK]",
    "key": "[KEY]",
    "pencil": "[EDIT]",
    "arrow_up": "^",
    "stop_sign": "[STOP]",
    "warning": "!",
    "locked": "[LOCK]",
    "pushpin": "[PIN]",
    "magnifying_glass_tilted_left": "[SCAN]",
    "fire": "[CRIT]",
    "yellow_circle": "[HIGH]",
    "sparkles": "*",
    "mag_right": "[VIEW]",
    "link": "->",
    "light_bulb": "[TIP]",
    "trophy": "[DONE]",
    "rocket": ">>",
    "busts_in_silhouette": "[TEAM]",
    "floppy_disk": "[SAVE]",
    "heavy_plus_sign": "[ADD]",
    "books": "[DOCS]",
    "speech_balloon": "[HELP]",
}

# Pre-compiled regex for emoji processing (Rich-style)
CUSTOM_EMOJI_PATTERN = re.compile(r"(:icon_\w+:)")


def process_custom_emojis(text: str, use_ascii: bool = False) -> str:
    """
    Pre-process our custom emoji namespace before Rich handles the text.
    This only handles our custom :icon_*: emojis.
    """
    if not isinstance(text, str) or ":icon_" not in text:
        return text

    def replace_custom_emoji(match: Match[str]) -> str:
        emoji_code = match.group(1)  # :icon_check:
        emoji_name = emoji_code[1:-1]  # icon_check

        # If we should use ASCII, use the fallback
        if use_ascii:
            return ASCII_FALLBACK_MAP.get(emoji_name, emoji_code)

        return CUSTOM_EMOJI_MAP.get(emoji_name, emoji_code)

    return CUSTOM_EMOJI_PATTERN.sub(replace_custom_emoji, text)


def process_rich_emojis_fallback(text: str) -> str:
    """
    Replace Rich emoji codes with ASCII alternatives when in problematic environments.
    """
    # Simple pattern to match Rich emoji codes like :emoji_name:
    emoji_pattern = re.compile(r":([a-zA-Z0-9_]+):")

    def replace_with_ascii(match: Match[str]) -> str:
        emoji_name = match.group(1)
        # Check if we have an ASCII fallback
        ascii_replacement = ASCII_FALLBACK_MAP.get(emoji_name, None)
        if ascii_replacement:
            return ascii_replacement

        # Otherwise keep the original
        return match.group(0)

    return emoji_pattern.sub(replace_with_ascii, text)


def load_emoji(text: str, use_ascii: bool = False) -> str:
    """
    Load emoji from text if emoji is present.
    """

    # Pre-process our custom emojis
    text = process_custom_emojis(text, use_ascii)

    # If we need ASCII fallbacks, also process Rich emoji codes
    if use_ascii:
        text = process_rich_emojis_fallback(text)

    return text



CMD_HELP_CODEBASE_INIT = "Initialize a Safety Codebase (like git init for security). Sets up a new codebase or connects your local project to an existing one on Safety Platform."
CMD_HELP_CODEBASE = (
    "[BETA] Manage your Safety Codebase integration.\nExample: safety codebase init"
)

CMD_CODEBASE_GROUP_NAME = "codebase"
CMD_CODEBASE_INIT_NAME = "init"


# init options help
CMD_HELP_CODEBASE_INIT_NAME = "Name of the codebase. Defaults to GIT origin name, parent directory name, or random string if parent directory is unnamed. The value will be normalized for use as an identifier."
CMD_HELP_CODEBASE_INIT_LINK_TO = (
    "Link to an existing codebase using its codebase slug (found in Safety Platform)."
)
CMD_HELP_CODEBASE_INIT_DISABLE_FIREWALL = "Don't enable Firewall protection for this codebase (enabled by default when available in your organization)"
CMD_HELP_CODEBASE_INIT_PATH = (
    "Path to the codebase directory. Defaults to current directory."
)


CODEBASE_INIT_REINITIALIZED = "Reinitialized existing codebase {codebase_name}"
CODEBASE_INIT_ALREADY_EXISTS = "A codebase already exists in this directory. Please delete .safety-project.ini and run `safety codebase init` again to initialize a new codebase."
CODEBASE_INIT_NOT_FOUND_LINK_TO = "\nError: codebase '{codebase_name}' specified with --link-to does not exist.\n\nTo create a new codebase instead, use one of:\n  safety codebase init\n  safety codebase init --name \"custom name\"\n\nTo link to an existing codebase, verify the codebase id and try again."
CODEBASE_INIT_NOT_FOUND_PROJECT_FILE = "\nError: codebase '{codebase_name}' specified with the current .safety-project.ini file does not exist.\n\nTo create a new codebase instead, delete the corrupted .safety-project.ini file and then use one of:\n  safety codebase init\n  safety codebase init --name \"custom name\"\n\nTo link to an existing codebase, verify the codebase id and try again."
CODEBASE_INIT_LINKED = "Linked to codebase {codebase_name}."
CODEBASE_INIT_CREATED = "Created new codebase {codebase_name}."
CODEBASE_INIT_ERROR = "Error: unable to initialize the codebase. Please try again."



import os
from pathlib import Path
import re
from tempfile import mkstemp
from typing import TYPE_CHECKING, List, Optional

import logging
import typer

from safety.models import ToolResult
from .parser import PipParser

from ..base import BaseCommand
from ..intents import ToolIntentionType
from safety_schemas.models.events.types import ToolType
from ..environment_diff import EnvironmentDiffTracker, PipEnvironmentDiffTracker
from ..mixins import InstallationAuditMixin
from ..utils import Pip
from ...encoding import detect_encoding


PIP_LOCK = "safety-pip.lock"

if TYPE_CHECKING:
    from ..environment_diff import EnvironmentDiffTracker

logger = logging.getLogger(__name__)


class PipCommand(BaseCommand):
    """
    Main class for hooks into pip commands.
    """

    def get_tool_type(self) -> ToolType:
        return ToolType.PIP

    def get_command_name(self) -> List[str]:
        """
        This uses command alias if available, with this we support
        pip3.13, pip3.12, etc.
        """

        cmd_name = ["pip"]

        if self._command_alias_used:
            cmd_name = [self._command_alias_used]

        return cmd_name

    def get_lock_path(self) -> str:
        return PIP_LOCK

    def get_diff_tracker(self) -> "EnvironmentDiffTracker":
        return PipEnvironmentDiffTracker()

    def should_track_state(self) -> bool:
        command_str = " ".join(self._args).lower()

        package_modifying_commands = [
            "install",
            "uninstall",
        ]

        return any(cmd in command_str for cmd in package_modifying_commands)

    @classmethod
    def from_args(cls, args: List[str], **kwargs):
        parser = PipParser()

        if intention := parser.parse(args):
            if intention.intention_type is ToolIntentionType.ADD_PACKAGE:
                return PipInstallCommand(args, intention=intention, **kwargs)

        return PipGenericCommand(args, **kwargs)


class PipGenericCommand(PipCommand):
    pass


class PipInstallCommand(PipCommand, InstallationAuditMixin):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._packages = []
        self.__index_url = None

    def before(self, ctx: typer.Context):
        super().before(ctx)
        args: List[Optional[str]] = self._args.copy()  # type: ignore

        if self._intention:
            for pkg in self._intention.packages:
                self._packages.append((pkg.name, pkg.version_constraint))

            if index_opt := self._intention.options.get(
                "index-url"
            ) or self._intention.options.get("i"):
                index_value = index_opt["value"]
                if index_value and index_value.startswith("https://pkgs.safetycli.com"):
                    self.__index_url = index_value

                arg_index = index_opt["arg_index"]
                value_index = index_opt["value_index"]

                if (
                    arg_index
                    and value_index
                    and arg_index < len(args)
                    and value_index < len(args)
                ):
                    args[arg_index] = None
                    args[value_index] = None

            if req_opt := self._intention.options.get(
                "requirement"
            ) or self._intention.options.get("r"):
                req_value = req_opt["value"]
                if req_value and Path(req_value).is_file():
                    with open(
                        req_value, "r", encoding=detect_encoding(Path(req_value))
                    ) as f:
                        fd, tmp_requirements_path = mkstemp(
                            suffix="safety-requirements.txt", text=True
                        )
                        with os.fdopen(fd, "w") as tf:
                            requirements = re.sub(
                                r"^(-i|--index-url).*$",
                                "",
                                f.read(),
                                flags=re.MULTILINE,
                            )
                            tf.write(requirements)

                        args[req_opt["value_index"]] = tmp_requirements_path

        self._args = [arg for arg in args if arg is not None]

    def after(self, ctx: typer.Context, result: ToolResult):
        super().after(ctx, result)
        self.handle_installation_audit(ctx, result)

    def env(self, ctx: typer.Context) -> dict:
        env = super().env(ctx)
        env["PIP_INDEX_URL"] = Pip.build_index_url(ctx, self.__index_url)
        return env



import configparser
import logging
from pathlib import Path
from dataclasses import dataclass
from typing import Optional
from safety_schemas.models import ProjectModel


PROJECT_CONFIG = ".safety-project.ini"
PROJECT_CONFIG_SECTION = "project"
PROJECT_CONFIG_ID = "id"
PROJECT_CONFIG_URL = "url"
PROJECT_CONFIG_NAME = "name"


logger = logging.getLogger(__name__)


@dataclass
class UnverifiedProjectModel:
    """
    Data class representing an unverified project model.
    """

    id: Optional[str]
    project_path: Path
    created: bool
    name: Optional[str] = None
    url_path: Optional[str] = None


def load_unverified_project_from_config(project_root: Path) -> UnverifiedProjectModel:
    """
    Loads an unverified project from the configuration file located at the project root.

    Args:
        project_root (Path): The root directory of the project.

    Returns:
        UnverifiedProjectModel: An instance of UnverifiedProjectModel.
    """
    config = configparser.ConfigParser()
    project_path = project_root / PROJECT_CONFIG
    config.read(project_path)
    id = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_ID, fallback=None)
    url = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_URL, fallback=None)
    name = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_NAME, fallback=None)
    created = True
    if not id:
        created = False

    return UnverifiedProjectModel(
        id=id, url_path=url, name=name, project_path=project_path, created=created
    )


def save_project_info(project: ProjectModel, project_path: Path) -> bool:
    """
    Saves the project information to the configuration file.

    Args:
        project (ProjectModel): The ProjectModel object containing project
                                information.
        project_path (Path): The path to the configuration file.

    Returns:
        bool: True if the project information was saved successfully, False
              otherwise.
    """
    config = configparser.ConfigParser()
    config.read(project_path)

    if PROJECT_CONFIG_SECTION not in config.sections():
        config[PROJECT_CONFIG_SECTION] = {}

    config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_ID] = project.id
    if project.url_path:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_URL] = project.url_path
    if project.name:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_NAME] = project.name

    try:
        with open(project_path, "w") as configfile:
            config.write(configfile)
    except Exception:
        logger.exception("Error saving project info")
        return False

    return True



# type: ignore
from unittest.mock import patch, MagicMock
from typing import List, Dict, Any

import typer

from safety.tool.base import BaseCommand


class CommandToolCaseMixin:
    """
    Generic test case for command implementations.
    """

    command_class = None
    command_args: List[str] = None
    parent_env: Dict[str, str] = None
    expected_env_vars: Dict[str, str] = None
    mock_configurations: List[Dict[str, Any]] = None

    def setUp(self):
        """
        Set up test environment before each test method.
        """
        if not self.command_class or not self.command_args:
            raise ValueError("command_class and command_args must be set by subclasses")

        self.command = self.command_class(self.command_args)
        self.parent_env = self.parent_env or {}
        self.expected_env_vars = self.expected_env_vars or {}
        self.mock_configurations = self.mock_configurations or []

    def test_env_preserves_existing_variables(self):
        """
        Test that env() method does not replace existing environment variables.
        """
        ctx = MagicMock(spec=typer.Context)

        # Setup parent environment
        existing_env = self.parent_env.copy()

        # Setup all required mocks
        mock_objects = {}
        for mock_config in self.mock_configurations:
            target = mock_config.get("target")
            return_value = mock_config.get("return_value")

            patcher = patch(target, return_value=return_value)
            mock = patcher.start()
            mock_objects[target] = mock
            self.addCleanup(patcher.stop)

        # Get the env from the command
        with patch.object(
            BaseCommand, "env", return_value=existing_env
        ) as mock_super_env:
            result_env = self.command.env(ctx)

            # Check that parent env variables are preserved
            for key, value in existing_env.items():
                self.assertEqual(result_env[key], value)

            # Check that expected env variables are added
            for key, value in self.expected_env_vars.items():
                self.assertIn(key, result_env)
                self.assertEqual(result_env[key], value)

            # Verify parent env was called
            mock_super_env.assert_called_once_with(ctx)



import logging
import os
import re
import shutil
from pathlib import Path
from sys import platform
from typing import TYPE_CHECKING, Dict, List

from .base import CommandInterceptor, Tool
from .types import InterceptorType

if TYPE_CHECKING or platform == "win32":
    import winreg

from typing import Union

logger = logging.getLogger(__name__)


class AutoRunManager:
    """
    Manages Windows Command Processor AutoRun registry entries
    """

    REGISTRY_KEY = r"Software\\Microsoft\\Command Processor"
    REGISTRY_VALUE = "AutoRun"

    def add_script(self, script_path: "Union[str, Path]") -> bool:
        """
        Add script to AutoRun, preserving existing commands
        """
        script_path = str(script_path)

        try:
            with self._open_registry_key() as key:
                tokens = self._get_current_tokens(key)

                if not self._script_exists_in_tokens(tokens, script_path):
                    new_tokens = (
                        [script_path, " & "] + tokens if tokens else [script_path]
                    )
                    self._set_autorun_value(key, "".join(new_tokens))

                return True
        except Exception:
            logger.info("Failed to add script to AutoRun")
            return False

    def remove_script(self, script_path: Union[str, Path]) -> bool:
        """
        Remove script from AutoRun, preserving other commands
        """
        script_path = str(script_path)

        try:
            with self._open_registry_key() as key:
                tokens = self._get_current_tokens(key)

                if self._script_exists_in_tokens(tokens, script_path):
                    cleaned_tokens = self._remove_script_tokens(tokens, script_path)

                    if cleaned_tokens:
                        self._set_autorun_value(key, " ".join(cleaned_tokens))
                    else:
                        self._delete_autorun_value(key)

                return True
        except Exception:
            logger.info("Failed to remove script from AutoRun")
            return False

    def get_current_commands(self) -> List[str]:
        """
        Get list of current AutoRun commands
        """
        try:
            with self._open_registry_key() as key:
                tokens = self._get_current_tokens(key)
                return [
                    token.strip()
                    for token in tokens
                    if not self._is_separator(token) and token.strip()
                ]
        except Exception:
            logger.info("Failed to get current AutoRun value")
            return []

    def _open_registry_key(self):
        """
        Context manager for registry key access
        """
        try:
            return winreg.OpenKey(
                winreg.HKEY_CURRENT_USER,
                self.REGISTRY_KEY,
                0,
                winreg.KEY_READ | winreg.KEY_SET_VALUE,
            )
        except FileNotFoundError:
            logger.info("Failed to open registry key")
            logger.info("Creating registry key: %s", self.REGISTRY_KEY)

            return winreg.CreateKey(winreg.HKEY_CURRENT_USER, self.REGISTRY_KEY)

    def _get_current_tokens(self, key) -> List[str]:
        """
        Get current AutoRun value as tokens
        """
        try:
            existing_value, _ = winreg.QueryValueEx(key, self.REGISTRY_VALUE)
            return self._tokenize_autorun(existing_value)
        except FileNotFoundError:
            logger.info("Failed to get current AutoRun value")
            return []

    def _is_our_script(self, token: str, script_path: str) -> bool:
        """
        Check if token is our script (ignoring whitespace)
        """
        return token.strip() == script_path

    def _is_separator(self, token: str) -> bool:
        """
        Check if token is a command separator that can be used to chain
        commands in the AutoRun value
        """
        return token.strip() in ["&", "&&", "|", "||"]

    def _tokenize_autorun(self, autorun_value: str) -> List[str]:
        """
        Tokenize AutoRun value preserving commands, separators, and spacing.
        Simple character-by-character parsing approach.
        """
        if not autorun_value:
            return []

        tokens = []
        current_token = ""
        i = 0

        while i < len(autorun_value):
            char = autorun_value[i]

            if char in "&|":
                # Save current token if exists
                if current_token:
                    tokens.append(current_token)
                    current_token = ""

                # Handle double operators (&&, ||)
                if i + 1 < len(autorun_value) and autorun_value[i + 1] == char:
                    tokens.append(char + char)  # && or ||
                    i += 2
                else:
                    tokens.append(char)  # & or |
                    i += 1
            else:
                current_token += char
                i += 1

        if current_token:
            tokens.append(current_token)

        return tokens

    def _script_exists_in_tokens(self, tokens: List[str], script_path: str) -> bool:
        """
        Check if script already exists in token list
        """
        return any(token.strip() == script_path for token in tokens)

    def _remove_script_tokens(self, tokens: List[str], script_path: str) -> List[str]:
        """
        Remove our script and clean up separators
        """
        result = []
        i = 0

        while i < len(tokens):
            if self._is_our_script(tokens[i], script_path):
                # Skip our script
                # Also skip the next separator if it exists
                if i + 1 < len(tokens) and self._is_separator(tokens[i + 1]):
                    i += 2  # Skip script + separator
                else:
                    i += 1  # Skip just the script
            else:
                result.append(tokens[i].strip())
                i += 1

        return result

    def _set_autorun_value(self, key, value: str):
        """
        Set AutoRun registry value
        """
        winreg.SetValueEx(key, self.REGISTRY_VALUE, 0, winreg.REG_SZ, value)

    def _delete_autorun_value(self, key):
        """
        Delete AutoRun registry value
        """
        winreg.DeleteValue(key, self.REGISTRY_VALUE)


class WindowsInterceptor(CommandInterceptor):
    def __init__(self):
        super().__init__(InterceptorType.WINDOWS_BAT)
        self.scripts_dir = Path.home() / "AppData" / "Local" / "safety"
        # Ensure the scripts directory exists
        # This makes sure that if a user is using a sandboxed Python
        # installation from the Microsoft Store, the directory is created and
        # the .resolve() method works correctly.
        self.scripts_dir.mkdir(parents=True, exist_ok=True)
        self.scripts_dir = self.scripts_dir.resolve()

        self.backup_dir = self.scripts_dir / "backups"
        self.backup_win_env_path = self.backup_dir / "path_backup.txt"
        self.venv_pwshell_wrapper_path = self.scripts_dir / "venv-wrappers.ps1"
        self.venv_cmd_wrapper_path = self.scripts_dir / "venv-wrappers.bat"

        # Update these markers could be a breaking change; be careful to handle
        # backward compatibility
        self.marker_start = ">>> Safety >>>"
        self.marker_end = "<<< Safety <<<"

    def _backup_path_env(self, path_content: str) -> None:
        """
        Backup current PATH to a file
        """
        self.backup_dir.mkdir(parents=True, exist_ok=True)

        metadata_lines = self._generate_metadata_content(prepend="")

        lines = (
            (self.marker_start,) + metadata_lines + (path_content,) + (self.marker_end,)
        )

        content = "\n".join(lines) + "\n"

        self.backup_win_env_path.write_text(content)

    def _generate_bat_content(self, tool_name: str) -> str:
        """
        Generate the content for the bat with metadata
        """
        metadata_lines = self._generate_metadata_content(prepend="REM")

        no_echo = "@echo off"
        wrapper = f"safety {tool_name} %*"
        lines = (
            (
                no_echo,
                f"REM {self.marker_start}",
            )
            + metadata_lines
            + (wrapper,)
            + (f"REM {self.marker_end}",)
        )

        return "\n".join(lines) + "\n"

    def __generate_cmd_wrapper_content(self, binaries: List[str]) -> str:
        """
        Generate the content for the cmd wrapper with metadata
        """
        metadata_lines = self._generate_metadata_content(prepend="REM")

        no_echo = "@echo off"
        wrappers = []
        for binary in binaries:
            wrapper = f"doskey {binary}={self.scripts_dir / binary}.bat $*"
            wrappers.append(wrapper)

        comment_top = f"{no_echo}\nREM {self.marker_start}"
        comment_bottom = f"REM {self.marker_end}"
        lines = (
            comment_top,
            *metadata_lines,
            *wrappers,
            comment_bottom,
        )

        return "\n".join(lines) + "\n"

    def __generate_powershell_wrapper_content(self, binaries: List[str]) -> str:
        """
        Generate the content for the powershell wrapper with PowerShell functions
        """
        metadata_lines = self._generate_metadata_content(prepend="#")

        wrappers = []
        for binary in binaries:
            bat_path = self.scripts_dir / f"{binary}.bat"
            wrapper = f"""function {binary} {{
        param([Parameter(ValueFromRemainingArguments)]$args)
        & "{bat_path}" @args
    }}"""
            wrappers.append(wrapper)

        comment_top = f"# {self.marker_start}"
        comment_bottom = f"# {self.marker_end}"
        lines = [
            comment_top,
            *metadata_lines,
            *wrappers,
            comment_bottom,
        ]

        return "\n".join(lines) + "\n"

    def __generate_powershell_sourcing_content(self, script_path: Path) -> str:
        """
        Generate the PowerShell sourcing content with Safety markers
        """
        lines = [
            f"# {self.marker_start}",
            f". {script_path}",
            f"# {self.marker_end}",
        ]

        return "\n".join(lines) + "\n"

    def __get_powershell_profiles(self) -> Dict[str, Path]:
        """
        Get the CurrentUserAllHosts profile paths for available PowerShell versions
        Returns a dictionary with shell executable as key and profile path as value
        """
        profiles = {}
        shells = [("powershell.exe", "PowerShell 5.1"), ("pwsh.exe", "PowerShell 7+")]

        for shell, shell_name in shells:
            try:
                import subprocess

                # Check if the shell is available
                try:
                    subprocess.run(
                        [shell, "-Command", "exit"],
                        capture_output=True,
                        text=True,
                        check=False,
                    )
                except FileNotFoundError:
                    logger.info(f"{shell_name} not found, skipping profile setup")
                    continue

                # Get the CurrentUserAllHosts profile path
                cmd = [
                    shell,
                    "-Command",
                    "Get-Variable PROFILE -ValueOnly | Select-Object -ExpandProperty CurrentUserAllHosts",
                ]
                result = subprocess.run(
                    cmd, capture_output=True, text=True, check=False
                )
                result_stdout = result.stdout.strip()
                if result.returncode == 0 and result_stdout:
                    profile_path = Path(result_stdout)
                    # Ensure parent directory exists
                    profile_path.parent.mkdir(parents=True, exist_ok=True)
                    # Create the file if it doesn't exist
                    if not profile_path.exists():
                        profile_path.touch()
                    profiles[shell] = profile_path
                    logger.info(f"Found {shell_name} profile at {profile_path}")
                else:
                    logger.info(
                        f"Failed to get {shell_name} profile path: {result.stderr.strip()}"
                    )
            except Exception as e:
                logger.info(f"Error while getting {shell_name} profile: {str(e)}")

        # Fallback to default profile path if no profiles were found
        if not profiles:
            default_path = (
                Path.home() / "Documents" / "WindowsPowerShell" / "profile.ps1"
            )
            default_path.parent.mkdir(parents=True, exist_ok=True)
            if not default_path.exists():
                default_path.touch()
            profiles["fallback"] = default_path
            logger.info(f"Using fallback profile at {default_path}")

        return profiles

    def _install_venv_wrappers(self, binaries: List[str]):
        """
        Install specific wrappers for virtualenv support on Windows
        """
        # Refresh scripts content
        # CMD wrappers
        cmd_wrapper = self.__generate_cmd_wrapper_content(binaries)
        self.venv_cmd_wrapper_path.write_text(cmd_wrapper)

        # PowerShell wrappers
        powershell_wrapper = self.__generate_powershell_wrapper_content(binaries)
        self.venv_pwshell_wrapper_path.write_text(powershell_wrapper)

        # Link CMD wrapper to Autorun
        autorun_manager = AutoRunManager()
        autorun_manager.add_script(self.venv_cmd_wrapper_path)

        # Link Powershell wrapper to Powershell PROFILEs
        profiles = self.__get_powershell_profiles()
        pwshell_source = self.__generate_powershell_sourcing_content(
            self.venv_pwshell_wrapper_path
        )

        for _, profile_path in profiles.items():
            try:
                # Read current content or create empty string if file doesn't exist yet
                try:
                    profile_content = profile_path.read_text()
                except FileNotFoundError:
                    profile_path.parent.mkdir(parents=True, exist_ok=True)
                    profile_content = ""

                # Add sourcing command if not already present
                if self.marker_start not in profile_content:
                    if profile_content and not profile_content.endswith("\n"):
                        profile_content += "\n"
                    profile_content += pwshell_source
                    profile_path.write_text(profile_content)
                    logger.info(f"Added PowerShell wrapper to {profile_path}")
            except Exception as e:
                logger.info(
                    f"Failed to update PowerShell profile at {profile_path}: {str(e)}"
                )

    def _remove_venv_wrappers(self):
        """
        Remove specific wrappers for virtualenv support on Windows.

        This is an indempotent operation.
        """
        # For CMD
        autorun_manager = AutoRunManager()
        autorun_manager.remove_script(self.venv_cmd_wrapper_path)

        # For PowerShell
        # Remove Powershell wrapper from all PowerShell profiles
        profiles = self.__get_powershell_profiles()

        for _, profile_path in profiles.items():
            try:
                if profile_path.exists():
                    profile_content = profile_path.read_text()

                    if self.marker_start not in profile_content:
                        logger.info(f"PowerShell wrapper not found in {profile_path}")
                        continue

                    # Look for our sourcing line and the comment block we added
                    # Remove the entire block including comments
                    lines = profile_content.splitlines()
                    new_lines = []
                    skip_block = False

                    for line in lines:
                        if self.marker_start in line:
                            skip_block = True
                            continue

                        if skip_block:
                            if self.marker_end in line:
                                skip_block = False
                            continue

                        new_lines.append(line)

                    new_content = "\n".join(new_lines)
                    new_content = re.sub(r"\n{3,}", "\n\n", new_content)

                    profile_path.write_text(new_content)
                    logger.info(f"Removed PowerShell wrapper from {profile_path}")
            except Exception as e:
                logger.info(
                    f"Failed to remove PowerShell wrapper from {profile_path}: {str(e)}"
                )

    def _batch_install_tools(self, tools: List[Tool]) -> bool:
        """
        Install interceptors for multiple tools at once
        """
        try:
            wrappers = []
            for tool in tools:
                for binary in tool.binary_names:
                    # TODO: Switch to binary once we support safety pip3, etc.
                    wrapper = self._generate_bat_content(tool.name)
                    wrappers.append((binary, wrapper))

            if not wrappers:
                return False

            # Create safety directory if it doesn't exist
            self.scripts_dir.mkdir(parents=True, exist_ok=True)

            for binary, wrapper in wrappers:
                wrapper_path = self.scripts_dir / f"{binary}.bat"
                wrapper_path.write_text(wrapper)

            # Virtualenv environment wrappers
            all_binaries = [binary for tool in tools for binary in tool.binary_names]
            self._install_venv_wrappers(binaries=all_binaries)

            # Add scripts directory to PATH if needed
            self._update_path()

            return True

        except Exception as e:
            logger.info("Failed to batch install tools: %s", e)
            return False

    def _batch_remove_tools(self, tools: List[Tool]) -> bool:
        """
        Remove interceptors for multiple tools at once.

        Note: We don't support removing specific tools yet,
        so we remove all tools.
        """
        try:
            self._update_path(remove=True)
            if self.scripts_dir.exists():
                shutil.rmtree(self.scripts_dir)

            self._remove_venv_wrappers()

            return True

        except Exception as e:
            logger.info("Failed to batch remove tools: %s", e)
            return False

    def _update_path(self, remove: bool = False) -> bool:
        """
        Update Windows PATH environment variable
        """

        try:
            with winreg.OpenKey(
                winreg.HKEY_CURRENT_USER, "Environment", 0, winreg.KEY_ALL_ACCESS
            ) as key:
                # Get current PATH value
                try:
                    path_val = winreg.QueryValueEx(key, "PATH")[0]
                    self._backup_path_env(path_content=path_val)
                except FileNotFoundError:
                    path_val = ""

                # Convert to Path objects
                paths = [Path(p) for p in path_val.split(os.pathsep) if p]

                if remove:
                    if self.scripts_dir in paths:
                        paths.remove(self.scripts_dir)
                        new_path = os.pathsep.join(str(p) for p in paths)
                        winreg.SetValueEx(
                            key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path
                        )
                else:
                    if self.scripts_dir not in paths:
                        paths.insert(0, self.scripts_dir)  # Add to beginning
                        new_path_val = os.pathsep.join(str(p) for p in paths)
                        winreg.SetValueEx(
                            key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path_val
                        )

            return True
        except Exception as e:
            logger.info("Failed to update PATH: %s", e)
            return False



from typing import Any, List, Protocol, Tuple, Dict, Optional, runtime_checkable
import typer
from rich.padding import Padding

from .base import EnvironmentDiffTracker

from safety.console import main_console as console
from safety.init.render import render_header, progressive_print
from safety.models import ToolResult
import logging

logger = logging.getLogger(__name__)


@runtime_checkable
class AuditableCommand(Protocol):
    """
    Protocol defining the contract for classes that can be audited for packages.
    """

    @property
    def _diff_tracker(self) -> "EnvironmentDiffTracker":
        """
        Provides package tracking functionality.
        """
        ...

    @property
    def _packages(self) -> List[Tuple[str, Optional[str]]]:
        """
        Provides the target package list.
        """
        ...


class InstallationAuditMixin:
    """
    Mixin providing installation audit functionality for command classes.

    This mixin can be used by any command class that needs to audit
    installation and show warnings.

    Classes using this mixin should conform to the AuditableCommand protocol.
    """

    def render_installation_warnings(
        self, ctx: typer.Context, packages_audit: Dict[str, Any]
    ):
        """
        Render installation warnings based on package audit results.

        Args:
            ctx: The typer context
            packages_audit: pre-fetched audit data
        """

        warning_messages = []
        for audited_package in packages_audit.get("audit", {}).get("packages", []):
            vulnerabilities = audited_package.get("vulnerabilities", {})
            critical_vulnerabilities = vulnerabilities.get("critical", 0)
            total_vulnerabilities = 0
            for count in vulnerabilities.values():
                total_vulnerabilities += count

            if total_vulnerabilities == 0:
                continue

            warning_message = f"[[yellow]Warning[/yellow]] {audited_package.get('package_specifier')} contains {total_vulnerabilities} vulnerabilities"
            if critical_vulnerabilities > 0:
                warning_message += f", including {critical_vulnerabilities} critical severity vulnerabilities"

            warning_message += "."
            warning_messages.append(warning_message)

        if len(warning_messages) > 0:
            console.print()
            render_header(" Safety Report")
            progressive_print(warning_messages)
            console.line()

    def render_package_details(self: "AuditableCommand"):
        """
        Render details for installed packages.
        """
        for package_name, _ in self._packages:
            console.print(
                Padding(
                    f"Learn more: [link]https://data.safetycli.com/packages/pypi/{package_name}/[/link]",
                    (0, 0, 0, 1),
                ),
                emoji=True,
            )

    def audit_packages(self, ctx: typer.Context) -> Dict[str, Any]:
        """
        Audit packages based on environment diff tracking.
        Override this method in your command class if needed.

        Args:
            ctx: The typer context

        Returns:
            Dict containing audit results
        """
        try:
            # Check if the instance has a diff tracker and can get a diff
            # Using getattr to avoid lint errors
            diff_tracker = getattr(self, "_diff_tracker", None)
            if diff_tracker and hasattr(diff_tracker, "get_diff"):
                added, _, updated = diff_tracker.get_diff()
                packages = {**added, **updated}

                if hasattr(ctx.obj, "auth") and hasattr(ctx.obj.auth, "client"):
                    return ctx.obj.auth.client.audit_packages(
                        [
                            f"{package_name}=={version[-1] if isinstance(version, tuple) else version}"
                            for (package_name, version) in packages.items()
                        ]
                    )
        except Exception:
            logger.debug("Audit API failed with error", exc_info=True)

        # Always return a dict to satisfy the return type
        return dict()

    def handle_installation_audit(self, ctx: typer.Context, result: ToolResult):
        """
        Handle installation audit and rendering warnings/details.
        This is an explicit method that can be called from a command's after method.

        Usage example:
            def after(self, ctx, result):
                super().after(ctx, result)
                self.handle_installation_audit(ctx, result)

        Args:
            ctx: The typer context
            result: The tool result
        """

        if not isinstance(self, AuditableCommand):
            raise TypeError(
                "handle_installation_audit can only be called on instances of AuditableCommand"
            )

        packages_audit = self.audit_packages(ctx)
        self.render_installation_warnings(ctx, packages_audit)

        # If command failed, show package details
        if not result.process or result.process.returncode != 0:
            self.render_package_details()



# type: ignore
import unittest
from safety.tool.uv.command import UvCommand
from ..base import CommandToolCaseMixin


class TestUvCommand(CommandToolCaseMixin, unittest.TestCase):
    """
    Test cases for UvCommand functionality.
    """

    command_class = UvCommand
    command_args = ["uv", "pip", "install", "package"]
    parent_env = {"EXISTING_VAR": "existing_value", "PATH": "/usr/bin:/bin"}
    expected_env_vars = {
        "UV_INDEX_SAFETY_USERNAME": "user",
        "UV_INDEX_SAFETY_PASSWORD": "mock_credentials_value",
    }
    mock_configurations = [
        {
            "target": "safety.tool.uv.command.index_credentials",
            "return_value": "mock_credentials_value",
        }
    ]



import unittest
import tempfile
from importlib.metadata import version
from click.testing import CliRunner
from packaging.version import Version
from safety.auth.models import Auth
from safety.cli import cli
from safety.console import main_console as console
from unittest.mock import patch


class TestScanCommand(unittest.TestCase):
    def setUp(self):
        # mix_stderr was removed in Click 8.2.0
        if Version(version("click")) >= Version("8.2.0"):
            self.runner = CliRunner()
        else:
            self.runner = CliRunner(mix_stderr=False)
        self.target = tempfile.mkdtemp()
        # Make sure the console is not quiet
        # TODO: This is a workaround, we should improve the way the console
        # is initialized in the CLI
        console.quiet = False

        cli.commands = cli.all_commands
        self.cli = cli

    @patch.object(Auth, "is_valid", return_value=False)
    @patch(
        "safety.auth.utils.SafetyAuthSession.get_authentication_type",
        return_value="unauthenticated",
    )
    def test_scan(self, mock_is_valid, mock_get_auth_type):
        result = self.runner.invoke(
            self.cli, ["scan", "--target", self.target, "--output", "json"]
        )
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(
            self.cli,
            [
                "--stage",
                "production",
                "scan",
                "--target",
                self.target,
                "--output",
                "json",
            ],
        )
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(
            self.cli,
            ["--stage", "cicd", "scan", "--target", self.target, "--output", "screen"],
        )
        self.assertEqual(result.exit_code, 1)



import logging
from pathlib import Path
import sys
import threading
import time
from typing import (
    TYPE_CHECKING,
    Optional,
    Iterator,
)

from rich.live import Live
from rich.padding import Padding
from rich.panel import Panel
from rich.prompt import Prompt
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from safety.codebase_utils import load_unverified_project_from_config
from safety.events.utils.emission import (
    emit_codebase_detection_status,
    emit_codebase_setup_completed,
    emit_init_exited,
    emit_init_scan_completed,
)
from safety.init.models import StepTracker
from safety_schemas.models.events.types import ToolType


from .render import (
    ask_codebase_setup,
    ask_continue,
    ask_firewall_setup,
    progressive_print,
    render_header,
    typed_print,
)
from safety.scan.init_scan import start_scan
from ..cli_util import (
    CommandType,
    FeatureType,
    SafetyCLICommand,
    SafetyCLISubGroup,
)
from safety.error_handlers import handle_cmd_exception
import typer


from safety.init.constants import (
    MSG_ANALYZE_CODEBASE_TITLE,
    MSG_CODEBASE_FAILED_TO_SCAN,
    MSG_CODEBASE_NOT_CONFIGURED,
    MSG_CODEBASE_URL_DESCRIPTION,
    MSG_COMPLETE_SECURED,
    MSG_COMPLETE_TOOL_SECURED,
    MSG_FIREWALL_UNINSTALL,
    MSG_LAST_MANUAL_STEP,
    MSG_NO_VULNERABILITIES_FOUND,
    MSG_NO_VULNS_CODEBASE_URL_DESCRIPTION,
    MSG_OPEN_DASHBOARD_PROMPT,
    MSG_SETUP_CODEBASE_NO_PROJECT,
    MSG_SETUP_COMPLETE_SUBTITLE,
    MSG_SETUP_COMPLETE_TITLE,
    MSG_SETUP_INCOMPLETE,
    MSG_SETUP_NEXT_STEPS,
    MSG_SETUP_NEXT_STEPS_MANUAL_STEP,
    MSG_SETUP_NEXT_STEPS_NO_PROJECT,
    MSG_SETUP_NEXT_STEPS_NO_VULNS,
    MSG_SETUP_NEXT_STEPS_SUBTITLE,
    MSG_SETUP_PACKAGE_FIREWALL_DESCRIPTION,
    MSG_SETUP_PACKAGE_FIREWALL_TITLE,
    MSG_SETUP_CODEBASE_DESCRIPTION,
    MSG_SETUP_CODEBASE_TITLE,
    CODEBASE_INIT_CMD_NAME,
    CODEBASE_INIT_HELP,
    CODEBASE_INIT_DIRECTORY_HELP,
    MSG_TOOLS_NOT_CONFIGURED,
    MSG_WELCOME_TITLE,
    MSG_WELCOME_DESCRIPTION,
)
from safety.init.main import create_project, launch_auth_if_needed, setup_firewall
from safety.console import (
    get_spinner_animation,
    main_console as console,
    should_use_ascii,
)
from ..tool.main import (
    configure_local_directory,
    find_local_tool_files,
)

from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE

from safety.decorators import notify
from safety.events.utils import emit_firewall_configured, emit_init_started
from safety_schemas.models.events.payloads import AliasConfig, IndexConfig, InitExitStep


if TYPE_CHECKING:
    import typer
    from safety.scan.init_scan import ScanResult

try:
    from typing import Annotated  # type: ignore
except ImportError:
    from typing_extensions import Annotated

init_app = typer.Typer(rich_markup_mode="rich", cls=SafetyCLISubGroup)

logger = logging.getLogger(__name__)


class InitScanState:
    """
    Class to track scan state for vulnerability scans

    Attributes:
        scan_id: ID of the scan
        dependencies: Number of dependencies found
        critical: Count of critical vulnerabilities
        high: Count of high severity vulnerabilities
        medium: Count of medium severity vulnerabilities
        low: Count of low severity vulnerabilities
        fixes: Number of fixes available
        fixed_vulns: Number of vulnerabilities with fixes
        url: URL to view the scan results
        completed: Whether the scan has completed
        progress: Percentage progress of the scan
        status_message: Current status message from the scanner
        status_action: Current action being performed by the scanner
        current_file: Current file being processed
    """

    def __init__(self):
        self.scan_id: Optional[str] = None
        self.dependencies: int = 0
        self.critical: int = 0
        self.high: int = 0
        self.medium: int = 0
        self.low: int = 0
        self.others: int = 0
        self.vulns_count: int = 0
        self.fixes: int = 0
        self.fixed_vulns: int = 0
        self.codebase_url: Optional[str] = None
        self.completed: bool = False
        self.progress: int = 0
        self.status_message: Optional[str] = None
        self.status_action: Optional[str] = None
        self.current_file: Optional[str] = None


def generate_summary(state: InitScanState, spinner_phase=0) -> Text:
    """
    Generate the summary text based on current scan state
    """
    spinner = get_spinner_animation()

    text_markup = f"Tested {state.dependencies} dependenc{'y' if state.dependencies == 1 else 'ies'} for security issues\n"
    text_markup += "\nFound:\n\n"

    categories = [
        {
            "name": "CRITICAL",
            "icon": ":fire:",
            "style": "bold red",
            "dim_style": "dim red",
            "count_attr": "critical",
            "spinner_offset": 0,
        },
        {
            "name": "HIGH",
            "icon": ":yellow_circle:",
            "style": "bold yellow",
            "dim_style": "dim yellow",
            "count_attr": "high",
            "spinner_offset": 2,
        },
        {
            "name": "MEDIUM",
            "icon": "!!",
            "style": "yellow",
            "dim_style": "dim yellow",
            "count_attr": "medium",
            "spinner_offset": 4,
        },
        {
            "name": "LOW",
            "icon": ":icon_info: ",
            "style": "bold blue",
            "dim_style": "dim blue",
            "count_attr": "low",
            "spinner_offset": 6,
        },
        {
            "name": "OTHER",
            "icon": "**",
            "style": "blue",
            "dim_style": "dim blue",
            "count_attr": "others",
            "spinner_offset": 8,
        },
    ]

    # No vulnerabilities case
    prepend_text_codebase_url = MSG_CODEBASE_URL_DESCRIPTION

    if state.completed and state.vulns_count <= 0:
        text_markup += MSG_NO_VULNERABILITIES_FOUND
        prepend_text_codebase_url = MSG_NO_VULNS_CODEBASE_URL_DESCRIPTION
    else:
        for category in categories:
            category_count = getattr(state, category["count_attr"])

            if state.completed and category_count == 0:
                continue

            style = category["style"]
            text_markup += f"[{style}]{category['icon']} {category['name']}: [/{style}]"

            if category_count > 0:
                vulns_word = f"vulnerabilit{'y' if category_count == 1 else 'ies'}"
                text_markup += f"[{style}]{category_count}[/{style}] {vulns_word}\n"
            else:
                spinner_status = spinner[
                    (spinner_phase + category["spinner_offset"]) % len(spinner)
                ]
                style = category["dim_style"]
                text_markup += f"[{style}]{spinner_status}[/{style}] Scanning\n"

        # Show fixes info if we have vulnerabilities
        if state.vulns_count > 0 and state.fixes is not None:
            text_markup += "\n"

            if state.fixes > 0:
                fix_word = f"fix{'' if state.fixes == 1 else 'es'}"
                vulns_word = f"vulnerabilit{'y' if state.fixed_vulns == 1 else 'ies'}"
                text_markup += f":sparkles: [green]{state.fixes}[/green] automatic {fix_word} available, resolving {state.fixed_vulns} {vulns_word}\n"
            else:
                text_markup += (
                    " No automatic fixes available for these vulnerabilities\n"
                )

    # Dashboard link if URL is available
    if state.codebase_url is not None:
        text_markup += f"\n{prepend_text_codebase_url}[blue underline]:link: {state.codebase_url}\n[/blue underline]"

    elif state.completed:
        text_markup += "\n"

    return console.render_str(text_markup)


def generate_status_updates(state: InitScanState, spinner_phase: int = 0) -> Text:
    """
    Generate text displaying current status updates and progress information

    Args:
        state: The InitScanState object containing status information
        spinner_phase: Current phase of the spinner animation

    Returns:
        Rich Text object containing formatted status updates
    """
    spinner = get_spinner_animation()

    markup_text = f"[cyan]{spinner[spinner_phase % len(spinner)]} Scanning...[/cyan]"

    # Display current status message if available
    if state.status_message:
        action_symbol = {
            "init": ":arrows_counterclockwise:",
            "scanning": ":magnifying_glass_tilted_left:",
            "uploading": ":cloud: ",
            "complete": ":white_heavy_check_mark:",
            "error": ":cross_mark:",
        }
        status_key = state.status_action if state.status_action is not None else "init"
        action_symbol = action_symbol.get(status_key, ":information_source: ")

        markup_text = f"\n[bold cyan]{action_symbol} Status: [/bold cyan][cyan]{state.status_message}[/cyan]\n"

        # If we're processing a file, show that
        if state.current_file and state.status_action == "scanning":
            markup_text += f"[bold cyan]:file_folder: Current file: [/bold cyan][dim cyan]{state.current_file}[/dim cyan]\n"

            # Ensure progress is capped at 100%
            display_progress = min(state.progress, 100)
            markup_text += f"[cyan]:bar_chart: Progress: {display_progress}%[/cyan]\n"

    return console.render_str(markup_text)


def process_scan_results(
    scan_results: Iterator["ScanResult"], state: InitScanState
) -> None:
    """Process the scan iterator and update state from typed result models

    Args:
        scan_results: Iterator yielding scan results from init_scan
        state: The InitScanState object to update with scan results
    """
    # Import the scan result types to handle typed results
    from safety.scan.init_scan import (
        ScanResultType,
    )

    try:
        for result in scan_results:
            # Now result is a typed model with proper attributes
            if result.type == ScanResultType.INIT:
                # Initial state with dependency count
                state.dependencies = result.dependencies
                state.status_message = "Initializing scan"
                state.status_action = "init"

            elif result.type == ScanResultType.PROGRESS:
                # Update all the state fields from the progress result
                # Type checker knows result is ProgressScanResult
                # Ensure progress never exceeds 100%
                state.progress = min(result.percent, 100)
                state.dependencies = result.dependencies

                # Track current file being processed
                state.current_file = result.file
                state.status_message = f"Processing {result.file}"
                state.status_action = "scanning"

                # Update severity counts if present
                if result.critical is not None:
                    state.critical = result.critical
                if result.high is not None:
                    state.high = result.high
                if result.medium is not None:
                    state.medium = result.medium
                if result.low is not None:
                    state.low = result.low
                if result.others is not None:
                    state.others = result.others

                # Update vulnerability count
                if result.vulns_count is not None:
                    state.vulns_count = result.vulns_count

                # Update fix information if present
                if result.fixes is not None:
                    state.fixes = result.fixes
                if result.fixed_vulns is not None:
                    state.fixed_vulns = result.fixed_vulns

            elif result.type == ScanResultType.STATUS:
                # Generic status update
                state.status_message = result.message
                state.status_action = result.action
                if result.percent is not None:
                    state.progress = min(result.percent, 100)

            elif result.type == ScanResultType.UPLOADING:
                # Status update for uploading phase
                state.status_message = result.message
                state.status_action = "uploading"
                if result.percent is not None:
                    state.progress = min(result.percent, 100)

            elif result.type == ScanResultType.COMPLETE:
                # Final update with complete data
                # Type checker knows result is CompleteScanResult
                state.progress = 100
                state.dependencies = result.dependencies
                state.critical = result.critical
                state.high = result.high
                state.medium = result.medium
                state.low = result.low
                state.others = result.others
                state.fixes = result.fixes
                state.fixed_vulns = result.fixed_vulns
                state.status_message = "Scan completed"
                state.status_action = "complete"
                state.vulns_count = result.vulns_count

                # Update project URL if available
                if result.codebase_url:
                    state.codebase_url = result.codebase_url

                if result.scan_id:
                    state.scan_id = result.scan_id

                # We're done processing
                state.completed = True

            # Add a small delay between updates to allow UI thread to refresh
            time.sleep(0.05)
    except Exception as e:
        console.print(f"Error processing scan results: {e}", style="bold red")
        state.status_message = f"Error: {str(e)}"
        state.status_action = "error"
    finally:
        # Ensure we mark as completed even if there was an exception
        state.completed = True


def init_scan_ui(ctx: "typer.Context", prompt_user: bool = False) -> InitScanState:
    """
    Initialize and run a scan for the init command, showing a live UI with scan progress.
    Uses the start_scan function to get an iterator of scan results and displays UI based on them.

    Args:
        ctx: The Typer context object containing configuration and project information
    """
    # Initialize state for tracking scan progress
    state = InitScanState()

    # Set up scan parameters and get the scan iterator
    target = ctx.obj.project.project_path.parent
    use_server_matching = False

    # Start the scan using the dedicated function
    scan_results = start_scan(
        ctx=ctx,
        target=target,
        use_server_matching=use_server_matching,
        auth_type=ctx.obj.auth.client.get_authentication_type(),
        is_authenticated=ctx.obj.auth.client.is_using_auth_credentials(),
        client=ctx.obj.auth.client,
        project=ctx.obj.project,
        platform_enabled=ctx.obj.platform_enabled,
    )

    # Process the scan results in a separate thread
    scan_thread = threading.Thread(
        target=process_scan_results, args=(scan_results, state)
    )
    scan_thread.daemon = True
    scan_thread.start()

    # Handle UI updates in the main thread
    spinner_phase = 0
    render_header(
        MSG_ANALYZE_CODEBASE_TITLE.format(project_name=ctx.obj.project.id),
        emoji=":magnifying_glass_tilted_left:",
    )
    time.sleep(0.8)

    # Detect if running on Windows
    is_windows = sys.platform == "win32" or should_use_ascii()

    # Alternate screen in Windows is buggy, so we disable it
    live_kwargs = {
        "refresh_per_second": 10,
        "screen": False if is_windows else True,
        "transient": False,
    }

    refresh_sleep = 0.1

    with Live(**live_kwargs) as live:
        while not state.completed or scan_thread.is_alive():
            # Update spinner phase for animation
            spinner_phase = (spinner_phase + 1) % 10

            # Summary information shown below status updates
            summary = generate_summary(state, spinner_phase)

            if is_windows:
                content = summary
            else:
                # Create a container for all UI elements
                container = Table.grid(padding=0, expand=True)
                container.add_row(None)
                container.add_row(
                    Panel(
                        generate_status_updates(state, spinner_phase),
                        border_style="cyan",
                        padding=(0, 1),
                    )
                )
                container.add_row(None)
                container.add_row(summary)
                content = container

            # Display the updated UI
            live.update(content)

            time.sleep(refresh_sleep)

        # Last sync
        if state.completed:
            live.update(generate_summary(state, spinner_phase))
        time.sleep(2)

    # Windows is not using alternate screen, so summary is already rendered
    if not is_windows:
        # Final update to ensure completion state is shown
        console.print(generate_summary(state))

    if state.codebase_url:
        typed_print(MSG_OPEN_DASHBOARD_PROMPT, end_line=False)
        should_open = "y"

        if prompt_user:
            should_open = Prompt.ask(
                "",
                choices=["y", "n", "Y", "N"],
                default="y",
                show_default=False,
                show_choices=False,
                console=console,
            ).lower()

        if should_open == "y":
            typer.launch(state.codebase_url)

    return state


@init_app.command(
    cls=SafetyCLICommand,
    help=CODEBASE_INIT_HELP,
    name=CODEBASE_INIT_CMD_NAME,
    options_metavar="[OPTIONS]",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
        CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL,
    },
)
@handle_cmd_exception
@notify
def init(
    ctx: typer.Context,
    directory: Annotated[
        Path,
        typer.Argument(  # type: ignore
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=True,
            show_default=False,
            help=CODEBASE_INIT_DIRECTORY_HELP,
        ),
    ] = Path("."),
):
    emit_init_started(ctx.obj.event_bus, ctx)
    # TODO: check if tty is available
    tracker = StepTracker()
    try:
        do_init(ctx, directory, tracker, prompt_user=console.is_interactive)
    except KeyboardInterrupt as e:
        emit_init_exited(ctx.obj.event_bus, ctx, exit_step=tracker.current_step)
        raise e


def do_init(
    ctx: typer.Context, directory: Path, tracker: StepTracker, prompt_user: bool = True
):
    """
    Initialize Safety CLI with the new onboarding flow.

    Args:
        ctx: The CLI context
        directory: The target directory to initialize
        prompt_user: Whether to prompt the user for input or use defaults
    """
    project_dir = directory.resolve()

    typed_print(MSG_WELCOME_TITLE)
    progressive_print(MSG_WELCOME_DESCRIPTION)

    tracker.current_step = InitExitStep.PRE_AUTH
    org_slug = launch_auth_if_needed(ctx, console)
    tracker.current_step = InitExitStep.POST_AUTH

    render_header(MSG_SETUP_PACKAGE_FIREWALL_TITLE, margin_right=1)
    console.print(MSG_SETUP_PACKAGE_FIREWALL_DESCRIPTION)

    console.print(
        Syntax(
            MSG_FIREWALL_UNINSTALL, "bash", theme="monokai", background_color="default"
        )
    )

    completed_tools = ""
    all_completed = False
    all_missing = True

    status = {
        ToolType.PIP: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
        ToolType.POETRY: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
        ToolType.UV: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
    }

    tracker.current_step = InitExitStep.PRE_FIREWALL_SETUP
    if ask_firewall_setup(ctx, prompt_user):
        completed_tools, all_completed, all_missing, status = setup_firewall(
            ctx, status, org_slug, console
        )
        console.line()
        ask_continue(ctx, prompt_user)
        console.line()

    tracker.current_step = InitExitStep.POST_FIREWALL_SETUP

    render_header(MSG_SETUP_CODEBASE_TITLE, emoji=":locked:")
    console.print(MSG_SETUP_CODEBASE_DESCRIPTION)

    project_scan_state = None

    tracker.current_step = InitExitStep.PRE_CODEBASE_SETUP
    local_files = find_local_tool_files(project_dir)

    emit_codebase_detection_status(
        event_bus=ctx.obj.event_bus,
        ctx=ctx,
        detected=any(local_files),
        detected_files=local_files if local_files else None,
    )

    if local_files:
        progressive_print(
            [
                f":pushpin: We found a `{file.name}` file in this directory."
                for file in local_files
            ]
        )

        console.line()

        if ask_codebase_setup(ctx, prompt_user):
            unverified_project = load_unverified_project_from_config(
                project_root=project_dir
            )
            link_behavior = "prompt"

            if unverified_project.created:
                link_behavior = "always"

            project_created, project_status = create_project(
                ctx,
                console,
                project_dir,
                unverified_project=unverified_project,
                link_behavior=link_behavior,
            )

            configure_local_directory(project_dir, org_slug, ctx.obj.project.id)

            emit_codebase_setup_completed(
                event_bus=ctx.obj.event_bus,
                ctx=ctx,
                is_created=project_created,
                codebase_id=ctx.obj.project.id if project_created else None,
            )

            if project_created:
                console.print(
                    "\n"
                    + f"{ctx.obj.project.id} codebase {project_status} :white_heavy_check_mark:"
                )
            else:
                progressive_print([":x: Failed to create codebase"])

            console.line()

            tracker.current_step = InitExitStep.PRE_SCAN
            project_scan_state = init_scan_ui(ctx, prompt_user)
            tracker.current_step = InitExitStep.POST_SCAN
            emit_init_scan_completed(
                event_bus=ctx.obj.event_bus,
                ctx=ctx,
                scan_id=project_scan_state.scan_id,
            )
    else:
        console.print(MSG_SETUP_CODEBASE_NO_PROJECT)

    tracker.current_step = InitExitStep.POST_CODEBASE_SETUP

    console.line()
    render_header(MSG_SETUP_COMPLETE_TITLE, emoji=":trophy:")

    is_setup_complete = all_completed and project_scan_state

    wrap_up_msg = []

    if all_completed:
        wrap_up_msg.append(
            MSG_COMPLETE_TOOL_SECURED.format(
                tools=completed_tools,
                firewall_url="https://platform.safetycli.com/firewall/",
            )
        )
    elif all_missing:
        wrap_up_msg.append(MSG_TOOLS_NOT_CONFIGURED)
    else:
        wrap_up_msg.append(MSG_SETUP_INCOMPLETE)

    wrap_up_msg.append("")

    if project_scan_state:
        if project_scan_state.scan_id:
            wrap_up_msg.append(
                MSG_COMPLETE_SECURED.format(
                    codebase_url=project_scan_state.codebase_url
                )
            )
        else:
            msg = (
                project_scan_state.status_message
                if project_scan_state.status_message
                else "Unknown"
            )
            wrap_up_msg.append(MSG_CODEBASE_FAILED_TO_SCAN.format(reason=msg))
    else:
        wrap_up_msg.append(MSG_CODEBASE_NOT_CONFIGURED)

    if wrap_up_msg:
        progressive_print(wrap_up_msg)
        console.line()

        if is_setup_complete:
            typed_print(MSG_SETUP_COMPLETE_SUBTITLE)
            console.line()
            typed_print(MSG_LAST_MANUAL_STEP)
            console.line()

    render_header(title=MSG_SETUP_NEXT_STEPS_SUBTITLE, emoji=":rocket:")
    console.line()

    next_steps_msg = MSG_SETUP_NEXT_STEPS

    if not project_scan_state:
        next_steps_msg = MSG_SETUP_NEXT_STEPS_NO_PROJECT
    elif project_scan_state.vulns_count <= 0:
        next_steps_msg = MSG_SETUP_NEXT_STEPS_NO_VULNS

    progressive_print(
        [Padding(console.render_str(line), (0, 0, 1, 0)) for line in next_steps_msg]
    )

    console.line()
    typed_print(MSG_SETUP_NEXT_STEPS_MANUAL_STEP, delay=0.04)
    console.line()

    # Emit event for firewall configuration
    emit_firewall_configured(
        event_bus=ctx.obj.event_bus,
        status=status,
    )
    tracker.current_step = InitExitStep.COMPLETED



import logging
import sys
import uuid
from rich.prompt import Prompt
import typer
from rich.console import Console

from safety.events.utils.emission import emit_firewall_setup_completed
from safety.init.render import progressive_print
from safety.util import clean_project_id

from ..tool import configure_system, configure_alias
from ..codebase_utils import load_unverified_project_from_config, save_project_info

from .constants import (
    MSG_AUTH_PROMPT,
    MSG_NEED_AUTHENTICATION,
    MSG_SETUP_INCOMPLETE,
    MSG_SETUP_PACKAGE_FIREWALL_NOTE_STATUS,
    MSG_SETUP_PACKAGE_FIREWALL_RESULT,
)

from pathlib import Path
from safety_schemas.models import ProjectModel
from safety_schemas.models.events.types import ToolType
from safety.scan.util import GIT
from ..auth.utils import SafetyAuthSession

from typing import TYPE_CHECKING, Any, Literal, Optional, Tuple
from safety.scan.render import (
    print_wait_project_verification,
    prompt_project_id,
    prompt_link_project,
)


if TYPE_CHECKING:
    from ..codebase_utils import UnverifiedProjectModel
    from ..models import SafetyCLI
    from .types import FirewallConfigStatus

logger = logging.getLogger(__name__)


def check_project(
    ctx: typer.Context,
    session: SafetyAuthSession,
    console: Console,
    unverified_project: "UnverifiedProjectModel",
    git_origin: Optional[str],
    ask_project_id: bool = False,
) -> dict:
    """
    Check the project against the session and stage, verifying the project if necessary.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        stage (Stage): The current stage.
        git_origin (Optional[str]): The Git origin URL.
        ask_project_id (bool): Whether to prompt for the project ID.

    Returns:
        dict: The result of the project check.
    """
    stage = ctx.obj.auth.stage
    source = ctx.obj.telemetry.safety_source if ctx.obj.telemetry else None
    data = {"scan_stage": stage, "safety_source": source}

    PRJ_SLUG_KEY = "project_slug"
    PRJ_SLUG_SOURCE_KEY = "project_slug_source"
    PRJ_GIT_ORIGIN_KEY = "git_origin"

    if git_origin:
        data[PRJ_GIT_ORIGIN_KEY] = git_origin

    if unverified_project.id:
        data[PRJ_SLUG_KEY] = unverified_project.id

        if unverified_project.created:
            data[PRJ_SLUG_SOURCE_KEY] = ".safety-project.ini"
        else:
            data[PRJ_SLUG_SOURCE_KEY] = "user"
    elif not git_origin or ask_project_id:
        fallback_id = unverified_project.project_path.parent.name

        if not fallback_id:
            # Sometimes the parent directory is empty, so we generate
            # a random ID
            fallback_id = str(uuid.uuid4())[:10]

        fallback_id = clean_project_id(fallback_id)

        if ask_project_id:
            id = prompt_project_id(console, fallback_id)
        else:
            id = fallback_id

        unverified_project.id = id

        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = "user"

    status = print_wait_project_verification(
        console,
        data[PRJ_SLUG_KEY] if data.get(PRJ_SLUG_KEY, None) else "-",
        (session.check_project, data),
        on_error_delay=1,
    )

    return status


def verify_project(
    console: Console,
    ctx: typer.Context,
    session: SafetyAuthSession,
    unverified_project: "UnverifiedProjectModel",
    git_origin: Optional[str],
    create_if_missing: bool = True,
    link_behavior: Literal["always", "prompt", "never"] = "prompt",
    prompt_for_name: bool = False,
) -> Tuple[bool, Optional[str]]:
    """
    Verify the project, linking it if necessary and saving the verified project information.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        git_origin (Optional[str]): The Git origin URL.
        create_if_missing (bool): Whether to create codebase if it doesn't exist. Defaults to True.
        link_behavior (Literal["always", "prompt", "never"]): How to handle codebase linking.
            - "always": Link without prompting
            - "prompt": Ask user before linking (default)
            - "never": Don't link, return early if codebase exists

    Returns:
        Tuple[bool, Optional[str]]: (success, status_message)
            - (True, "created"): New codebase was created and verified
            - (True, "linked"): Existing codebase was linked
            - (True, "found"): Codebase found but not linked (link_behavior="never")
            - (False, "not_found"): Codebase not found and create_if_missing=False
            - (False, None): Verification failed
    """

    # Track if we need to ask for project ID (when user declines linking)
    ask_for_project_id = False
    asked_codebase_asked_before = False

    while True:
        result = check_project(
            ctx,
            session,
            console,
            unverified_project,
            git_origin,
            # Ask for project ID when:
            # 1. User previously declined linking and we need a new ID
            # 2. We're in prompt mode and don't have a project ID yet
            ask_project_id=ask_for_project_id,
        )

        unverified_slug = result.get("slug")
        project = result.get("project", None)

        # Handle case where project doesn't exist
        if not project:
            if not create_if_missing:
                return (False, "not_found")
            # Project will be created - continue to verification
            project_status = (True, "created")

            if prompt_for_name and not asked_codebase_asked_before:
                asked_codebase_asked_before = True
                ask_for_project_id = True
                continue
        else:
            # Project exists - handle based on link_behavior
            if link_behavior == "never":
                unverified_project.id = project.get("slug")
                return (True, "found")
            elif link_behavior == "always":
                project_status = (True, "linked")
            elif link_behavior == "prompt":
                # Prompt user for confirmation
                prj_name = project.get("name", None)
                prj_admin_email = project.get("admin", None)

                should_link = prompt_link_project(
                    prj_name=prj_name, prj_admin_email=prj_admin_email, console=console
                )

                if should_link:
                    project_status = (True, "linked")
                else:
                    # User declined linking, ask for new project ID and retry
                    unverified_project.id = None
                    ask_for_project_id = True
                    continue

        # Proceed with project verification
        verified_prj = print_wait_project_verification(
            console,
            unverified_slug,  # type: ignore
            (session.project, {"project_id": unverified_slug}),
            on_error_delay=1,
        )

        if (
            verified_prj
            and isinstance(verified_prj, dict)
            and verified_prj.get("slug", None)
        ):
            save_verified_project(
                ctx,
                verified_prj["slug"],
                verified_prj.get("name", None),
                unverified_project.project_path,
                verified_prj.get("url", None),
                verified_prj.get("organization", None),
            )
            return project_status
        else:
            # Verification failed
            return (False, None)


def save_verified_project(
    ctx: typer.Context,
    slug: str,
    name: Optional[str],
    project_path: Path,
    url_path: Optional[str],
    organization: Optional[dict],
):
    """
    Save the verified project information to the context and project info file.

    Args:
        ctx (typer.Context): The context of the Typer command.
        slug (str): The project slug.
        name (Optional[str]): The project name.
        project_path (Path): The project path.
        url_path (Optional[str]): The project URL path.
        organization (Optional[str]): The project organization.
    """
    ctx.obj.project = ProjectModel(
        id=slug, name=name, project_path=project_path, url_path=url_path
    )

    save_project_info(project=ctx.obj.project, project_path=project_path)

    ctx.obj.org = {}
    if organization:
        ctx.obj.org = {
            "name": organization.get("name"),
            "slug": organization.get("slug"),
        }


def create_project(
    ctx: typer.Context,
    console: Console,
    target: Path,
    unverified_project: Optional["UnverifiedProjectModel"] = None,
    create_if_missing: bool = True,
    link_behavior: Literal["always", "prompt", "never"] = "prompt",
) -> Tuple[bool, Optional[str]]:
    """
    Loads existing project from the specified target locations or creates a new project.

    Args:
        ctx: The CLI context
        console: The console object
        target (Path): The target location
        unverified_project (UnverifiedProjectModel): The unverified project model
        create_if_missing (bool): Whether to create codebase if it doesn't exist
        link_behavior (Literal["always", "prompt", "never"]): How to handle codebase linking
    """
    # Load .safety-project.ini
    if not unverified_project:
        unverified_project = load_unverified_project_from_config(project_root=target)

    session = ctx.obj.auth.client
    git_data = GIT(root=target).build_git_data()
    origin = None

    if git_data:
        origin = git_data.origin

    if ctx.obj.platform_enabled:
        result = verify_project(
            console,
            ctx,
            session,
            unverified_project,
            origin,
            create_if_missing,
            link_behavior,
        )
        if ctx.obj.project:
            ctx.obj.project.git = git_data
        return result
    else:
        console.print("Project creation is not supported for your account.")
        return (False, None)


def launch_auth_if_needed(ctx: typer.Context, console: Console) -> Optional[str]:
    """
    Launch the authentication flow if needed.

    Args:
        ctx: The CLI context

    Returns:
        Optional[str]: The organization slug if authentication is successful
    """
    obj: "SafetyCLI" = ctx.obj
    org_slug = None

    if (
        not obj.auth
        or not obj.auth.client
        or not obj.auth.client.is_using_auth_credentials()
    ):
        console.print(MSG_NEED_AUTHENTICATION)

        if not console.is_interactive:
            sys.exit(0)

        auth_choice = Prompt.ask(
            MSG_AUTH_PROMPT,
            choices=["r", "l", "R", "L"],
            default="L",
            show_choices=False,
            show_default=True,
            console=console,
        ).lower()

        from safety.auth.cli import auth_app
        from safety.cli_util import get_command_for

        login_command = get_command_for(name="login", typer_instance=auth_app)
        register_command = get_command_for(name="register", typer_instance=auth_app)

        ctx.obj.only_auth_msg = True

        if auth_choice == "r":
            ctx.invoke(register_command)
        else:
            ctx.invoke(login_command)

    try:
        data = ctx.obj.auth.client.initialize()
        org_slug = data.get("organization-data", {}).get("slug")
    except Exception:
        logger.exception("Unable to load data on the init command")

    return org_slug


def setup_firewall(
    ctx: Any, status: "FirewallConfigStatus", org_slug: Optional[str], console: Console
) -> Tuple[str, bool, bool, "FirewallConfigStatus"]:
    """
    Setup the firewall, this function also handles the output.

    Args:
        ctx: The CLI context
        status: The current status of the firewall
        org_slug: The organization slug
        console: The console object

    Returns:
        Tuple[bool, bool, FirewallConfigStatus]: A tuple containing the following:
            - bool: True if all tools are configured, False otherwise
            - bool: True if all tools are missing, False otherwise
            - FirewallConfigStatus: The current status of the firewall
    """
    emoji_check = "[green]:icon_check:[/green]"

    configured_index = configure_system(org_slug)
    configured_alias = configure_alias()
    if configured_alias is None:
        configured_alias = []

    console.line()

    configured = {}
    if configured_index:
        configured["index"] = configured_index

    if configured_alias:
        configured["alias"] = configured_alias

    if any([item[1] for item in configured_index]) or any(
        [item[1] for item in configured_alias]
    ):
        for config_type, results in configured.items():
            for tool_type, path in results:
                tool_name = tool_type.value
                index_type = "global"

                tool_config = status[tool_type]
                is_configured = False

                if path:
                    if config_type == "index":
                        msg = f"Configured {tool_name}’s {index_type} index"
                    else:
                        msg = f"Aliased {tool_name} to safety"

                    is_configured = True
                    configured_msg = f"{emoji_check} {msg}"

                    path = path.resolve()

                    if len(path.parts) > 1:
                        progressive_print([f"{configured_msg} (`{path}`)"])
                    else:
                        progressive_print([configured_msg])
                else:
                    if config_type == "index":
                        msg = f"{tool_name}’s {index_type} index"
                    else:
                        msg = f"{tool_name} alias"

                    prefix_msg = "Failed to configure"
                    emoji = "[red bold]x[/red bold]"

                    # If there is a non-compatible global index
                    if tool_type in [ToolType.POETRY]:
                        prefix_msg = "Skipped"
                        msg += " - not supported by poetry"
                        emoji = "[gray bold]-[/gray bold]"
                        # TODO: Set None for now, to avoid mixing
                        # no configured error with skipped.
                        tool_config[config_type] = None
                    else:
                        is_configured = False

                    progressive_print([f"{emoji} {prefix_msg} {msg}"])

                if config_obj := tool_config[config_type]:
                    config_obj.is_configured = is_configured

        console.line()
    else:
        progressive_print(["[red bold]x[/red bold] Failed to configure system"])

    completed = []
    missing = []
    for tool_type, tool_status in status.items():
        for config_type, config_obj in tool_status.items():
            if config_obj:
                if config_obj.is_configured:
                    completed.append(config_obj)
                else:
                    missing.append(config_obj)

    all_completed = not missing
    all_missing = not completed

    tools = [tool_type.value.title() for tool_type in status]
    completed_tools = (
        ", ".join(tools[:-1]) + " and " + tools[-1] if len(tools) > 1 else tools[0]
    )

    if all_completed:
        console.print(
            f"{emoji_check} {completed_tools} {MSG_SETUP_PACKAGE_FIREWALL_RESULT}"
        )
        console.print(MSG_SETUP_PACKAGE_FIREWALL_NOTE_STATUS)
    else:
        progressive_print([MSG_SETUP_INCOMPLETE])

    console.line()

    emit_firewall_setup_completed(
        event_bus=ctx.obj.event_bus,
        ctx=ctx,
        status=status,
    )

    return completed_tools, all_completed, all_missing, status



import logging
from pathlib import Path
from typing import Optional
from safety.codebase.render import render_initialization_result
from safety.errors import SafetyError
from safety.events.utils.emission import (
    emit_codebase_detection_status,
    emit_codebase_setup_completed,
)
from safety.init.main import launch_auth_if_needed
from safety.tool.main import find_local_tool_files
from safety.util import clean_project_id
from typing_extensions import Annotated

import typer
from safety.cli_util import SafetyCLISubGroup, SafetyCLICommand
from .constants import (
    CMD_CODEBASE_INIT_NAME,
    CMD_HELP_CODEBASE_INIT,
    CMD_HELP_CODEBASE,
    CMD_CODEBASE_GROUP_NAME,
    CMD_HELP_CODEBASE_INIT_DISABLE_FIREWALL,
    CMD_HELP_CODEBASE_INIT_LINK_TO,
    CMD_HELP_CODEBASE_INIT_NAME,
    CMD_HELP_CODEBASE_INIT_PATH,
)

from ..cli_util import CommandType, get_command_for
from ..error_handlers import handle_cmd_exception
from ..decorators import notify
from ..constants import CONTEXT_COMMAND_TYPE, DEFAULT_EPILOG
from safety.console import main_console as console
from .main import initialize_codebase, prepare_unverified_codebase


logger = logging.getLogger(__name__)

cli_apps_opts = {
    "rich_markup_mode": "rich",
    "cls": SafetyCLISubGroup,
    "name": CMD_CODEBASE_GROUP_NAME,
}
codebase_app = typer.Typer(**cli_apps_opts)

DEFAULT_CMD = CMD_CODEBASE_INIT_NAME


@codebase_app.callback(
    invoke_without_command=True,
    cls=SafetyCLISubGroup,
    help=CMD_HELP_CODEBASE,
    epilog=DEFAULT_EPILOG,
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
    },
)
def codebase(
    ctx: typer.Context,
):
    """
    Group command for Safety Codebase (project) operations. Running this command will forward to the default command.
    """
    logger.info("codebase started")

    # If no subcommand is invoked, forward to the default command
    if not ctx.invoked_subcommand:
        default_command = get_command_for(name=DEFAULT_CMD, typer_instance=codebase_app)
        return ctx.forward(default_command)


@codebase_app.command(
    cls=SafetyCLICommand,
    help=CMD_HELP_CODEBASE_INIT,
    name=CMD_CODEBASE_INIT_NAME,
    epilog=DEFAULT_EPILOG,
    options_metavar="[OPTIONS]",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
@handle_cmd_exception
@notify
def init(
    ctx: typer.Context,
    name: Annotated[
        Optional[str],
        typer.Option(
            help=CMD_HELP_CODEBASE_INIT_NAME,
            callback=lambda name: clean_project_id(name) if name else None,
        ),
    ] = None,
    link_to: Annotated[
        Optional[str],
        typer.Option(
            "--link-to",
            help=CMD_HELP_CODEBASE_INIT_LINK_TO,
            callback=lambda name: clean_project_id(name) if name else None,
        ),
    ] = None,
    skip_firewall_setup: Annotated[
        bool, typer.Option(help=CMD_HELP_CODEBASE_INIT_DISABLE_FIREWALL)
    ] = False,
    codebase_path: Annotated[
        Path,
        typer.Option(
            "--path",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
            show_default=False,
            help=CMD_HELP_CODEBASE_INIT_PATH,
        ),
    ] = Path("."),
):
    """
    Initialize a Safety Codebase. The codebase may be entirely new to Safety Platform,
    or may already exist in Safety Platform and the user is wanting to initialize it locally.
    """
    logger.info("codebase init started")

    if link_to and name:
        raise typer.BadParameter("--link-to and --name cannot be used together")

    org_slug = launch_auth_if_needed(ctx, console)

    if not org_slug:
        raise SafetyError(
            "Organization not found, please run 'safety auth status' or 'safety auth login'"
        )

    should_enable_firewall = not skip_firewall_setup and ctx.obj.firewall_enabled

    unverified_codebase = prepare_unverified_codebase(
        codebase_path=codebase_path,
        user_provided_name=name,
        user_provided_link_to=link_to,
    )

    local_files = find_local_tool_files(codebase_path)

    emit_codebase_detection_status(
        event_bus=ctx.obj.event_bus,
        ctx=ctx,
        detected=any(local_files),
        detected_files=local_files if local_files else None,
    )

    project_file_created, project_status = initialize_codebase(
        ctx=ctx,
        console=console,
        codebase_path=codebase_path,
        unverified_codebase=unverified_codebase,
        org_slug=org_slug,
        link_to=link_to,
        should_enable_firewall=should_enable_firewall,
    )

    codebase_init_status = (
        "reinitialized" if unverified_codebase.created else project_status
    )
    codebase_id = ctx.obj.project.id if ctx.obj.project and ctx.obj.project.id else None

    render_initialization_result(
        console=console,
        codebase_init_status=codebase_init_status,
        codebase_id=codebase_id,
    )

    emit_codebase_setup_completed(
        event_bus=ctx.obj.event_bus,
        ctx=ctx,
        is_created=project_file_created,
        codebase_id=codebase_id,
    )



from functools import lru_cache
import logging
import os
import sys
from typing import TYPE_CHECKING, List, Dict, Any, Optional, Union
from rich.console import Console
from rich.theme import Theme
from safety.emoji import load_emoji


if TYPE_CHECKING:
    from rich.console import HighlighterType, JustifyMethod, OverflowMethod
    from rich.style import Style
    from rich.text import Text


LOG = logging.getLogger(__name__)


@lru_cache()
def should_use_ascii():
    """
    Check if we should use ASCII alternatives for emojis
    """
    encoding = getattr(sys.stdout, "encoding", "").lower()

    if encoding in {"utf-8", "utf8", "cp65001", "utf-8-sig"}:
        return False

    return True


def get_spinner_animation() -> List[str]:
    """
    Get the spinner animation based on the encoding
    """
    if should_use_ascii():
        spinner = [
            "[    ]",
            "[=   ]",
            "[==  ]",
            "[=== ]",
            "[====]",
            "[ ===]",
            "[  ==]",
            "[   =]",
        ]
    else:
        spinner = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
    return spinner


def replace_non_ascii_chars(text: str):
    """
    Replace non-ascii characters with ascii alternatives
    """
    CHARS_MAP = {
        "━": "-",
        "’": "'",
    }

    for char, replacement in CHARS_MAP.items():
        text = text.replace(char, replacement)

    try:
        text.encode("ascii")
    except UnicodeEncodeError:
        LOG.warning("No handled non-ascii characters detected, encoding with replace")
        text = text.encode("ascii", "replace").decode("ascii")

    return text


class SafeConsole(Console):
    """
    Console subclass that handles emoji encoding issues by detecting
    problematic encoding environments and replacing emojis with ASCII alternatives.
    Uses string replacement for custom emoji namespace to avoid private API usage.
    """

    def render_str(
        self,
        text: str,
        *,
        style: Union[str, "Style"] = "",
        justify: Optional["JustifyMethod"] = None,
        overflow: Optional["OverflowMethod"] = None,
        emoji: Optional[bool] = None,
        markup: Optional[bool] = None,
        highlight: Optional[bool] = None,
        highlighter: Optional["HighlighterType"] = None,
    ) -> "Text":
        """
        Override render_str to pre-process our custom emojis before Rich handles the text.
        """

        use_ascii = should_use_ascii()
        text = load_emoji(text, use_ascii=use_ascii)

        if use_ascii:
            text = replace_non_ascii_chars(text)

        # Let Rich handle everything else normally
        return super().render_str(
            text,
            style=style,
            justify=justify,
            overflow=overflow,
            emoji=emoji,
            markup=markup,
            highlight=highlight,
            highlighter=highlighter,
        )


SAFETY_THEME = {
    "file_title": "bold default on default",
    "dep_name": "bold yellow on default",
    "scan_meta_title": "bold default on default",
    "vuln_brief": "red on default",
    "rem_brief": "bold green on default",
    "rem_severity": "bold red on default",
    "brief_severity": "bold default on default",
    "status.spinner": "green",
    "recommended_ver": "bold cyan on default",
    "vuln_id": "bold default on default",
    "number": "bold cyan on default",
    "link": "underline bright_blue on default",
    "tip": "bold default on default",
    "specifier": "bold cyan on default",
    "vulns_found_number": "red on default",
}


non_interactive = os.getenv("NON_INTERACTIVE") == "1"

console_kwargs: Dict[str, Any] = {
    "theme": Theme(SAFETY_THEME, inherit=False),
    "emoji": not should_use_ascii(),
}

if non_interactive:
    LOG.info(
        "NON_INTERACTIVE environment variable is set, forcing non-interactive mode"
    )
    console_kwargs["force_terminal"] = True
    console_kwargs["force_interactive"] = False

main_console = SafeConsole(**console_kwargs)



# type: ignore

import pytest
from unittest.mock import MagicMock, patch

import typer

from safety.tool.pip.command import PipInstallCommand
from safety.tool.uv.command import UvInstallCommand
from safety.tool.poetry.command import PoetryAddCommand


class TestInstallationCommandsAudit:
    """
    Test suite for verifying installation audit functionality in command classes.
    """

    def setup_method(self):
        """
        Set up test fixtures.
        """
        self.ctx = MagicMock(spec=typer.Context)
        self.ctx.obj = MagicMock()
        self.result = MagicMock(duration_ms=100, process=MagicMock(returncode=0))

    @pytest.mark.parametrize(
        "command_class,command_args",
        [
            (PipInstallCommand, ["install", "requests"]),
            (UvInstallCommand, ["pip", "install", "requests"]),
            (PoetryAddCommand, ["add", "requests"]),
        ],
    )
    @patch("safety.tool.base.BaseCommand._handle_command_result")
    def test_installation_command_calls_audit(
        self, mock_handle_result, command_class, command_args
    ):
        """
        Test that all installation commands call handle_installation_audit in after().
        """
        command = command_class(command_args)

        with patch.object(
            command_class, "handle_installation_audit"
        ) as mock_handle_audit:
            command.after(self.ctx, self.result)

            mock_handle_audit.assert_called_once_with(self.ctx, self.result)



from pathlib import Path
import logging

logger = logging.getLogger(__name__)


def detect_encoding(file_path: Path) -> str:
    """
    UTF-8 is the most common encoding standard, this is a simple
    way to improve the support for related Windows based files.

    Handles the most common cases efficiently.
    """
    try:
        with open(file_path, "rb") as f:
            # Read first 3 bytes for BOM detection
            bom = f.read(3)

        # Check most common Windows patterns first
        if bom[:2] in (b"\xff\xfe", b"\xfe\xff"):
            return "utf-16"
        elif bom.startswith(b"\xef\xbb\xbf"):
            return "utf-8-sig"

        return "utf-8"
    except Exception:
        logger.exception("Error detecting encoding")
        return "utf-8"
